# Copyright 2024 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class EmptytreeReinstallUnsatisfiabilityTestCase(TestCase):
    def testEmptytreeReinstallUnsatisfiability(self):
        """
        Tests to check if emerge fails and complains when --emptytree
        package dependency graph reinstall is unsatisfied, even if the already
        installed packages successfully satisfy the dependency tree.

        See bug #651018 where emerge silently skips package
        reinstalls because of unsatisfied use flag requirements.
        """
        ebuilds = {
            "dev-libs/A-1": {
                "DEPEND": "dev-libs/B",
                "RDEPEND": "dev-libs/B",
                "EAPI": "2",
            },
            "dev-libs/B-1": {
                "DEPEND": "dev-libs/C[foo]",
                "RDEPEND": "dev-libs/C[foo]",
                "EAPI": "2",
            },
            "dev-libs/C-1": {
                "IUSE": "foo",
                "EAPI": "2",
            },
            "dev-libs/X-1": {
                "DEPEND": "dev-libs/Y[-baz]",
                "RDEPEND": "dev-libs/Y[-baz]",
                "EAPI": "2",
            },
            "dev-libs/Y-1": {
                "IUSE": "baz",
                "EAPI": "2",
            },
            "dev-libs/Z-1": {
                "DEPEND": "dev-libs/W",
                "RDEPEND": "dev-libs/W",
                "EAPI": "2",
            },
            "dev-libs/W-1": {
                "EAPI": "2",
            },
        }

        installed = {
            "dev-libs/A-1": {
                "DEPEND": "dev-libs/B",
                "RDEPEND": "dev-libs/B",
                "EAPI": "2",
            },
            "dev-libs/B-1": {
                "DEPEND": "dev-libs/C[foo]",
                "RDEPEND": "dev-libs/C[foo]",
                "EAPI": "2",
            },
            "dev-libs/C-1": {
                "IUSE": "foo",
                "USE": "foo",
                "EAPI": "2",
            },
            "dev-libs/X-1": {
                "DEPEND": "dev-libs/Y[-baz]",
                "RDEPEND": "dev-libs/Y[-baz]",
                "EAPI": "2",
            },
            "dev-libs/Y-1": {
                "IUSE": "baz",
                "USE": "-baz",
                "EAPI": "2",
            },
            "dev-libs/Z-1": {
                "DEPEND": "dev-libs/W",
                "RDEPEND": "dev-libs/W",
                "EAPI": "2",
            },
            "dev-libs/W-1": {
                "EAPI": "2",
            },
        }

        user_config = {
            "package.use": ("dev-libs/Y baz",),
            "package.mask": ("dev-libs/W",),
        }

        world = ["dev-libs/X"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--emptytree": True},
                success=False,
                mergelist=["dev-libs/C-1", "dev-libs/B-1", "dev-libs/A-1"],
                use_changes={"dev-libs/C-1": {"foo": True}},
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--emptytree": True, "--exclude": ["dev-libs/C"]},
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-1"],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--emptytree": True},
                success=False,
                mergelist=["dev-libs/Y-1", "dev-libs/X-1"],
                use_changes={"dev-libs/Y-1": {"baz": False}},
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/Z"],
                options={"--emptytree": True},
                success=False,
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            user_config=user_config,
            world=world,
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# archive_conf.py -- functionality common to archive-conf and dispatch-conf
# Copyright 2003-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

# Library by Wayne Davison <gentoo@blorf.net>, derived from code
# written by Jeremy Wohl (http://igmus.org)

import errno
import functools
import shlex
import stat
import subprocess
import sys
import tempfile

import portage
from portage import _encodings, os, shutil
from portage.env.loaders import KeyValuePairFileLoader
from portage.localization import _
from portage.util import varexpand
from portage.util.hooks import perform_hooks
from portage.util.path import iter_parents

RCS_BRANCH = "1.1.1"
RCS_LOCK = "rcs -ko -M -l"
RCS_PUT = 'ci -t-"Archived config file." -m"dispatch-conf update."'
RCS_GET = "co"
_ARCHIVE_ROTATE_MAX = 9


def diffstatusoutput(cmd, file1, file2):
    """
    Execute the string cmd in a shell with getstatusoutput() and return a
    2-tuple (status, output).
    """
    # Use Popen to emulate getstatusoutput(), since getstatusoutput() may
    # raise a UnicodeDecodeError which makes the output inaccessible.
    args = shlex.split(cmd % (file1, file2))

    args = (portage._unicode_encode(x, errors="strict") for x in args)
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output = portage._unicode_decode(proc.communicate()[0])
    if output and output.endswith("\n"):
        # getstatusoutput strips one newline
        output = output[:-1]
    return (proc.wait(), output)


def diff_mixed(func, file1, file2):
    tempdir = None
    try:
        if (
            os.path.islink(file1)
            and not os.path.islink(file2)
            and os.path.isfile(file1)
            and os.path.isfile(file2)
        ):
            # If a regular file replaces a symlink to a regular
            # file, then show the diff between the regular files
            # (bug #330221).
            diff_files = (file2, file2)
        else:
            files = [file1, file2]
            diff_files = [file1, file2]
            for i in range(len(diff_files)):
                try:
                    st = os.lstat(diff_files[i])
                except OSError:
                    st = None
                if st is not None and stat.S_ISREG(st.st_mode):
                    continue

                if tempdir is None:
                    tempdir = tempfile.mkdtemp()
                diff_files[i] = os.path.join(tempdir, f"{i}")
                if st is None:
                    content = "/dev/null\n"
                elif stat.S_ISLNK(st.st_mode):
                    link_dest = os.readlink(files[i])
                    content = f"SYM: {file1} -> {link_dest}\n"
                elif stat.S_ISDIR(st.st_mode):
                    content = f"DIR: {file1}\n"
                elif stat.S_ISFIFO(st.st_mode):
                    content = f"FIF: {file1}\n"
                else:
                    content = f"DEV: {file1}\n"
                with open(diff_files[i], mode="w", encoding=_encodings["stdio"]) as f:
                    f.write(content)

        return func(diff_files[0], diff_files[1])

    finally:
        if tempdir is not None:
            shutil.rmtree(tempdir)


class diff_mixed_wrapper:
    def __init__(self, f, *args):
        self._func = f
        self._args = args

    def __call__(self, *args):
        return diff_mixed(
            functools.partial(self._func, *(self._args + args[:-2])), *args[-2:]
        )


diffstatusoutput_mixed = diff_mixed_wrapper(diffstatusoutput)


def read_config(mandatory_opts):
    eprefix = portage.settings["EPREFIX"]
    if portage._not_installed:
        config_path = os.path.join(
            portage.PORTAGE_BASE_PATH, "cnf", "dispatch-conf.conf"
        )
    else:
        config_path = os.path.join(eprefix or os.sep, "etc/dispatch-conf.conf")
    loader = KeyValuePairFileLoader(config_path, None)
    opts, _errors = loader.load()
    if not opts:
        print(_(f"dispatch-conf: Error reading {config_path}; fatal"), file=sys.stderr)
        sys.exit(1)

    # Handle quote removal here, since KeyValuePairFileLoader doesn't do that.
    quotes = "\"'"
    for k, v in opts.items():
        if v[:1] in quotes and v[:1] == v[-1:]:
            opts[k] = v[1:-1]

    for key in mandatory_opts:
        if key not in opts:
            if key == "merge":
                opts["merge"] = "sdiff --suppress-common-lines --output='%s' '%s' '%s'"
            else:
                print(
                    _(
                        f'dispatch-conf: Missing option "{key}" in /etc/dispatch-conf.conf; fatal'
                    ),
                    file=sys.stderr,
                )

    # archive-dir supports ${EPREFIX} expansion, in order to avoid hardcoding
    variables = {"EPREFIX": eprefix}
    opts["archive-dir"] = varexpand(opts["archive-dir"], mydict=variables)

    if not os.path.exists(opts["archive-dir"]):
        os.mkdir(opts["archive-dir"])
        # Use restrictive permissions by default, in order to protect
        # against vulnerabilities (like bug #315603 involving rcs).
        os.chmod(opts["archive-dir"], 0o700)
    elif not os.path.isdir(opts["archive-dir"]):
        print(
            _(
                rf"""dispatch-conf: Config archive dir [{opts["archive-dir"]}] must exist; fatal"""
            ),
            file=sys.stderr,
        )
        sys.exit(1)

    return opts


def _archive_copy(src_st, src_path, dest_path):
    """
    Copy file from src_path to dest_path. Regular files and symlinks
    are supported. If an EnvironmentError occurs, then it is logged
    to stderr.

    @param src_st: source file lstat result
    @type src_st: posix.stat_result
    @param src_path: source file path
    @type src_path: str
    @param dest_path: destination file path
    @type dest_path: str
    """
    # Remove destination file in order to ensure that the following
    # symlink or copy2 call won't fail (see bug #535850).
    try:
        os.unlink(dest_path)
    except OSError:
        pass
    try:
        if stat.S_ISLNK(src_st.st_mode):
            os.symlink(os.readlink(src_path), dest_path)
        else:
            shutil.copy2(src_path, dest_path)
    except OSError as e:
        portage.util.writemsg(
            f"dispatch-conf: Error copying {src_path} to {dest_path}: {e}\n",
            noiselevel=-1,
        )


def rcs_archive(archive, curconf, newconf, mrgconf):
    """Archive existing config in rcs (on trunk). Then, if mrgconf is
    specified and an old branch version exists, merge the user's changes
    and the distributed changes and put the result into mrgconf.  Lastly,
    if newconf was specified, leave it in the archive dir with a .dist.new
    suffix along with the last 1.1.1 branch version with a .dist suffix."""

    try:
        os.makedirs(os.path.dirname(archive))
    except OSError:
        pass

    try:
        curconf_st = os.lstat(curconf)
    except OSError:
        curconf_st = None

    if curconf_st is not None and (
        stat.S_ISREG(curconf_st.st_mode) or stat.S_ISLNK(curconf_st.st_mode)
    ):
        _archive_copy(curconf_st, curconf, archive)

    if os.path.lexists(f"{archive},v"):
        os.system(f"{RCS_LOCK} {archive}")
    os.system(f"{RCS_PUT} {archive}")

    ret = 0
    mystat = None
    if newconf:
        try:
            mystat = os.lstat(newconf)
        except OSError:
            pass

    if mystat is not None and (
        stat.S_ISREG(mystat.st_mode) or stat.S_ISLNK(mystat.st_mode)
    ):
        os.system(f"{RCS_GET} -r{RCS_BRANCH} {archive}")
        has_branch = os.path.lexists(archive)
        if has_branch:
            os.rename(archive, f"{archive}.dist")

        _archive_copy(mystat, newconf, archive)

        if (
            has_branch
            and mrgconf
            and os.path.isfile(archive)
            and os.path.isfile(mrgconf)
        ):
            # This puts the results of the merge into mrgconf.
            ret = os.system(f"rcsmerge -p -r{RCS_BRANCH} '{archive}' > '{mrgconf}'")
            os.chmod(mrgconf, mystat.st_mode)
            os.chown(mrgconf, mystat.st_uid, mystat.st_gid)
        os.rename(archive, f"{archive}.dist.new")

    return ret


def _file_archive_rotate(archive):
    """
    Rename archive to archive + '.1', and perform similar rotation
    for files up to archive + '.9'.

    @param archive: file path to archive
    @type archive: str
    """

    max_suf = 0
    try:
        for max_suf, max_st, max_path in (
            (suf, os.lstat(path), path)
            for suf, path in (
                (suf, f"{archive}.{suf}") for suf in range(1, _ARCHIVE_ROTATE_MAX + 1)
            )
        ):
            pass
    except OSError as e:
        if e.errno not in (errno.ENOENT, errno.ESTALE):
            raise
        # There's already an unused suffix.
    else:
        # Free the max suffix in order to avoid possible problems
        # when we rename another file or directory to the same
        # location (see bug 256376).
        if stat.S_ISDIR(max_st.st_mode):
            # Removing a directory might destroy something important,
            # so rename it instead.
            head, tail = os.path.split(archive)
            placeholder = tempfile.NamedTemporaryFile(prefix=f"{tail}.", dir=head)
            placeholder.close()
            os.rename(max_path, placeholder.name)
        else:
            os.unlink(max_path)

        # The max suffix is now unused.
        max_suf -= 1

    for suf in range(max_suf + 1, 1, -1):
        os.rename(f"{archive}.{suf - 1}", f"{archive}.{suf}")

    os.rename(archive, f"{archive}.1")


def _file_archive_ensure_dir(parent_dir):
    """
    Ensure that the parent directory for an archive exists.
    If a file exists where a directory is needed, then rename
    it (see bug 256376).

    @param parent_dir: path of parent directory
    @type parent_dir: str
    """

    for parent in iter_parents(parent_dir):
        # Use lstat because a symlink to a directory might point
        # to a directory outside of the config archive, making
        # it an unsuitable parent.
        try:
            parent_st = os.lstat(parent)
        except OSError:
            pass
        else:
            if not stat.S_ISDIR(parent_st.st_mode):
                _file_archive_rotate(parent)
            break

    try:
        os.makedirs(parent_dir)
    except OSError:
        pass


def file_archive(archive, curconf, newconf, mrgconf):
    """Archive existing config to the archive-dir, bumping old versions
    out of the way into .# versions (log-rotate style). Then, if mrgconf
    was specified and there is a .dist version, merge the user's changes
    and the distributed changes and put the result into mrgconf.  Lastly,
    if newconf was specified, archive it as a .dist.new version (which
    gets moved to the .dist version at the end of the processing)."""

    _file_archive_ensure_dir(os.path.dirname(archive))

    # Archive the current config file if it isn't already saved
    if (
        os.path.lexists(archive)
        and len(diffstatusoutput_mixed("diff -aq '%s' '%s'", curconf, archive)[1]) != 0
    ):
        _file_archive_rotate(archive)

    try:
        curconf_st = os.lstat(curconf)
    except OSError:
        curconf_st = None

    if curconf_st is not None and (
        stat.S_ISREG(curconf_st.st_mode) or stat.S_ISLNK(curconf_st.st_mode)
    ):
        _archive_copy(curconf_st, curconf, archive)

    mystat = None
    if newconf:
        try:
            mystat = os.lstat(newconf)
        except OSError:
            pass

    if mystat is not None and (
        stat.S_ISREG(mystat.st_mode) or stat.S_ISLNK(mystat.st_mode)
    ):
        # Save off new config file in the archive dir with .dist.new suffix
        newconf_archive = f"{archive}.dist.new"
        if os.path.isdir(newconf_archive) and not os.path.islink(newconf_archive):
            _file_archive_rotate(newconf_archive)
        _archive_copy(mystat, newconf, newconf_archive)

        ret = 0
        if (
            mrgconf
            and os.path.isfile(curconf)
            and os.path.isfile(newconf)
            and os.path.isfile(f"{archive}.dist")
        ):
            # This puts the results of the merge into mrgconf.
            ret = os.system(
                f"diff3 -mE '{curconf}' '{archive}.dist' '{newconf}' > '{mrgconf}'"
            )
            os.chmod(mrgconf, mystat.st_mode)
            os.chown(mrgconf, mystat.st_uid, mystat.st_gid)

        return ret


def rcs_archive_post_process(archive):
    """Check in the archive file with the .dist.new suffix on the branch
    and remove the one with the .dist suffix."""
    os.rename(f"{archive}.dist.new", archive)
    if os.path.lexists(f"{archive}.dist"):
        # Commit the last-distributed version onto the branch.
        os.system(f"{RCS_LOCK}{RCS_BRANCH} {archive}")
        os.system(f"{RCS_PUT} -r{RCS_BRANCH} {archive}")
        os.unlink(f"{archive}.dist")
    else:
        # Forcefully commit the last-distributed version onto the branch.
        os.system(f"{RCS_PUT} -f -r{RCS_BRANCH} {archive}")


def file_archive_post_process(archive):
    """Rename the archive file with the .dist.new suffix to a .dist suffix"""
    if os.path.lexists(f"{archive}.dist.new"):
        dest = f"{archive}.dist"
        if os.path.isdir(dest) and not os.path.islink(dest):
            _file_archive_rotate(dest)
        os.rename(f"{archive}.dist.new", dest)


def perform_conf_update_hooks(kind, conf):
    """Invoke the hooks in the conf-update.d directory. The notification
    'kind' must either be 'pre-update' or 'post-update'. And 'conf' is the
    absolute path to the configuration file that is going to be or was
    updated."""
    perform_hooks("conf-update.d", kind, conf)


def perform_conf_update_session_hooks(kind):
    """Invoke the hooks in the conf-update-session.d directory. The
    notification 'kind' must either be 'pre-session' or 'post-session'."""
    perform_hooks("conf-update.d", kind)



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotCollisionTestCase(TestCase):
    def testSlotCollision(self):
        ebuilds = {
            "dev-libs/A-1": {"PDEPEND": "foo? ( dev-libs/B )", "IUSE": "foo"},
            "dev-libs/B-1": {"IUSE": "foo"},
            "dev-libs/C-1": {"DEPEND": "dev-libs/A[foo]", "EAPI": 2},
            "dev-libs/D-1": {
                "DEPEND": "dev-libs/A[foo=] dev-libs/B[foo=]",
                "IUSE": "foo",
                "EAPI": 2,
            },
            "dev-libs/E-1": {},
            "dev-libs/E-2": {"IUSE": "foo"},
            "app-misc/Z-1": {},
            "app-misc/Z-2": {},
            "app-misc/Y-1": {"DEPEND": "=app-misc/Z-1"},
            "app-misc/Y-2": {"DEPEND": ">app-misc/Z-1"},
            "app-misc/X-1": {"DEPEND": "=app-misc/Z-2"},
            "app-misc/X-2": {"DEPEND": "<app-misc/Z-2"},
            "sci-libs/K-1": {"IUSE": "+foo", "EAPI": 1},
            "sci-libs/L-1": {"DEPEND": "sci-libs/K[-foo]", "EAPI": 2},
            "sci-libs/M-1": {"DEPEND": "sci-libs/K[foo=]", "IUSE": "+foo", "EAPI": 2},
            "sci-libs/Q-1": {"SLOT": "1", "IUSE": "+bar foo", "EAPI": 1},
            "sci-libs/Q-2": {
                "SLOT": "2",
                "IUSE": "+bar +foo",
                "EAPI": 2,
                "PDEPEND": "sci-libs/Q:1[bar?,foo?]",
            },
            "sci-libs/P-1": {"DEPEND": "sci-libs/Q:1[foo=]", "IUSE": "foo", "EAPI": 2},
            "sys-libs/A-1": {
                "RDEPEND": "foo? ( sys-libs/J[foo=] )",
                "IUSE": "+foo",
                "EAPI": "4",
            },
            "sys-libs/B-1": {
                "RDEPEND": "bar? ( sys-libs/J[bar=] )",
                "IUSE": "+bar",
                "EAPI": "4",
            },
            "sys-libs/C-1": {"RDEPEND": "sys-libs/J[bar]", "EAPI": "4"},
            "sys-libs/D-1": {"RDEPEND": "sys-libs/J[bar?]", "IUSE": "bar", "EAPI": "4"},
            "sys-libs/E-1": {
                "RDEPEND": "sys-libs/J[foo(+)?]",
                "IUSE": "+foo",
                "EAPI": "4",
            },
            "sys-libs/F-1": {"RDEPEND": "sys-libs/J[foo(+)]", "EAPI": "4"},
            "sys-libs/J-1": {"IUSE": "+foo", "EAPI": "4"},
            "sys-libs/J-2": {"IUSE": "+bar", "EAPI": "4"},
            "app-misc/A-1": {
                "IUSE": "foo +bar",
                "REQUIRED_USE": "^^ ( foo bar )",
                "EAPI": "4",
            },
            "app-misc/B-1": {"DEPEND": "=app-misc/A-1[foo=]", "IUSE": "foo", "EAPI": 2},
            "app-misc/C-1": {"DEPEND": "=app-misc/A-1[foo]", "EAPI": 2},
            "app-misc/E-1": {"RDEPEND": "dev-libs/E[foo?]", "IUSE": "foo", "EAPI": "2"},
            "app-misc/F-1": {"RDEPEND": "=dev-libs/E-1", "IUSE": "foo", "EAPI": "2"},
            "dev-lang/perl-5.12": {"SLOT": "0/5.12", "EAPI": "5"},
            "dev-lang/perl-5.16": {"SLOT": "0/5.16", "EAPI": "5"},
        }
        installed = {
            "dev-libs/A-1": {
                "PDEPEND": "foo? ( dev-libs/B )",
                "IUSE": "foo",
                "USE": "foo",
            },
            "dev-libs/B-1": {"IUSE": "foo", "USE": "foo"},
            "dev-libs/C-1": {"DEPEND": "dev-libs/A[foo]", "EAPI": 2},
            "dev-libs/D-1": {
                "DEPEND": "dev-libs/A[foo=] dev-libs/B[foo=]",
                "IUSE": "foo",
                "USE": "foo",
                "EAPI": 2,
            },
            "sci-libs/K-1": {"IUSE": "foo", "USE": ""},
            "sci-libs/L-1": {"DEPEND": "sci-libs/K[-foo]"},
            "sci-libs/Q-1": {
                "SLOT": "1",
                "IUSE": "+bar +foo",
                "USE": "bar foo",
                "EAPI": 1,
            },
            "sci-libs/Q-2": {
                "SLOT": "2",
                "IUSE": "+bar +foo",
                "USE": "bar foo",
                "EAPI": 2,
                "PDEPEND": "sci-libs/Q:1[bar?,foo?]",
            },
            "app-misc/A-1": {
                "IUSE": "+foo bar",
                "USE": "foo",
                "REQUIRED_USE": "^^ ( foo bar )",
                "EAPI": "4",
            },
        }

        test_cases = (
            # A qt-*[qt3support] like mess.
            ResolverPlaygroundTestCase(
                ["dev-libs/A", "dev-libs/B", "dev-libs/C", "dev-libs/D"],
                options={"--autounmask": "n"},
                success=False,
                mergelist=[
                    "dev-libs/A-1",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                slot_collision_solutions=[
                    {"dev-libs/A-1": {"foo": True}, "dev-libs/D-1": {"foo": True}}
                ],
            ),
            ResolverPlaygroundTestCase(
                [
                    "sys-libs/A",
                    "sys-libs/B",
                    "sys-libs/C",
                    "sys-libs/D",
                    "sys-libs/E",
                    "sys-libs/F",
                ],
                options={"--autounmask": "n"},
                success=False,
                ignore_mergelist_order=True,
                slot_collision_solutions=[],
                mergelist=[
                    "sys-libs/J-2",
                    "sys-libs/J-1",
                    "sys-libs/A-1",
                    "sys-libs/B-1",
                    "sys-libs/C-1",
                    "sys-libs/D-1",
                    "sys-libs/E-1",
                    "sys-libs/F-1",
                ],
            ),
            # A version based conflicts, nothing we can do.
            ResolverPlaygroundTestCase(
                ["=app-misc/X-1", "=app-misc/Y-1"],
                success=False,
                mergelist=[
                    "app-misc/Z-1",
                    "app-misc/Z-2",
                    "app-misc/X-1",
                    "app-misc/Y-1",
                ],
                ignore_mergelist_order=True,
                slot_collision_solutions=[],
            ),
            ResolverPlaygroundTestCase(
                ["=app-misc/X-2", "=app-misc/Y-2"],
                success=False,
                mergelist=[
                    "app-misc/Z-1",
                    "app-misc/Z-2",
                    "app-misc/X-2",
                    "app-misc/Y-2",
                ],
                ignore_mergelist_order=True,
                slot_collision_solutions=[],
            ),
            ResolverPlaygroundTestCase(
                ["=app-misc/E-1", "=app-misc/F-1"],
                success=False,
                mergelist=[
                    "dev-libs/E-1",
                    "dev-libs/E-2",
                    "app-misc/E-1",
                    "app-misc/F-1",
                ],
                ignore_mergelist_order=True,
                slot_collision_solutions=[],
            ),
            # sub-slot
            ResolverPlaygroundTestCase(
                [
                    "dev-lang/perl:0/5.12",
                    "dev-lang/perl:0/5.16",
                    "=dev-lang/perl-5.12*",
                ],
                success=False,
                mergelist=["dev-lang/perl-5.12", "dev-lang/perl-5.16"],
                ignore_mergelist_order=True,
                slot_collision_solutions=[],
            ),
            # Simple cases.
            ResolverPlaygroundTestCase(
                ["sci-libs/L", "sci-libs/M"],
                success=False,
                mergelist=["sci-libs/L-1", "sci-libs/M-1", "sci-libs/K-1"],
                ignore_mergelist_order=True,
                slot_collision_solutions=[
                    {"sci-libs/K-1": {"foo": False}, "sci-libs/M-1": {"foo": False}}
                ],
            ),
            # Avoid duplicates.
            ResolverPlaygroundTestCase(
                ["sci-libs/P", "sci-libs/Q:2"],
                success=False,
                options={
                    "--update": True,
                    "--complete-graph": True,
                    "--autounmask": "n",
                },
                mergelist=["sci-libs/P-1", "sci-libs/Q-1"],
                ignore_mergelist_order=True,
                all_permutations=True,
                slot_collision_solutions=[
                    {"sci-libs/Q-1": {"foo": True}, "sci-libs/P-1": {"foo": True}}
                ],
            ),
        )
        # NOTE: For this test case, ResolverPlaygroundTestCase attributes
        # vary randomly between runs, so it's expected to fail randomly.
        # Conflict with REQUIRED_USE
        # ResolverPlaygroundTestCase(
        # 	["=app-misc/C-1", "=app-misc/B-1"],
        # 	all_permutations = True,
        # 	slot_collision_solutions = None,
        # 	use_changes={"app-misc/A-1": {"foo": True}},
        # 	mergelist = ["app-misc/A-1", "app-misc/C-1", "app-misc/B-1"],
        # 	ignore_mergelist_order = True,
        # 	success = False),
        # )

        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testConnectedCollision(self):
        """
        Ensure that we are able to solve connected slot conflicts
        which cannot be solved each on their own.
        """
        ebuilds = {
            "dev-libs/A-1": {"RDEPEND": "=dev-libs/X-1"},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/X"},
            "dev-libs/X-1": {"RDEPEND": "=dev-libs/Y-1"},
            "dev-libs/X-2": {"RDEPEND": "=dev-libs/Y-2"},
            "dev-libs/Y-1": {"PDEPEND": "=dev-libs/X-1"},
            "dev-libs/Y-2": {"PDEPEND": "=dev-libs/X-2"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A", "dev-libs/B"],
                all_permutations=True,
                options={"--backtrack": 0},
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    "dev-libs/Y-1",
                    "dev-libs/X-1",
                    ("dev-libs/A-1", "dev-libs/B-1"),
                ],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, debug=False)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testDeeplyConnectedCollision(self):
        """
        Like testConnectedCollision, except that there is another
        level of dependencies between the two conflicts.
        """
        ebuilds = {
            "dev-libs/A-1": {"RDEPEND": "=dev-libs/X-1"},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/X"},
            "dev-libs/X-1": {"RDEPEND": "dev-libs/K"},
            "dev-libs/X-2": {"RDEPEND": "dev-libs/L"},
            "dev-libs/K-1": {"RDEPEND": "=dev-libs/Y-1"},
            "dev-libs/L-1": {"RDEPEND": "=dev-libs/Y-2"},
            "dev-libs/Y-1": {"PDEPEND": "=dev-libs/X-1"},
            "dev-libs/Y-2": {"PDEPEND": "=dev-libs/X-2"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A", "dev-libs/B"],
                all_permutations=True,
                options={"--backtrack": 0},
                success=True,
                ignore_mergelist_order=True,
                mergelist=[
                    "dev-libs/Y-1",
                    "dev-libs/X-1",
                    "dev-libs/K-1",
                    "dev-libs/A-1",
                    "dev-libs/B-1",
                ],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, debug=False)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testSelfDEPENDRemovalCrash(self):
        """
        Make sure we don't try to remove a packages twice. This happened
        in the past when a package had a DEPEND on itself.
        """
        ebuilds = {
            "dev-libs/A-1": {"RDEPEND": "=dev-libs/X-1"},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/X"},
            "dev-libs/X-1": {},
            "dev-libs/X-2": {"DEPEND": ">=dev-libs/X-2"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A", "dev-libs/B"],
                all_permutations=True,
                success=True,
                ignore_mergelist_order=True,
                mergelist=["dev-libs/X-1", "dev-libs/A-1", "dev-libs/B-1"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, debug=False)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



#!/usr/bin/env python
# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import os
import sys

from portage.util import apply_recursive_permissions

# Change back to original cwd _after_ all imports (bug #469338).
os.chdir(os.environ["__PORTAGE_HELPER_CWD"])


def main(files):
    # We can't trust that the filesystem encoding (locale dependent)
    # correctly matches the arguments, so use surrogateescape to
    # pass through the original argv bytes for Python 3.
    fs_encoding = sys.getfilesystemencoding()
    files = [x.encode(fs_encoding, "surrogateescape") for x in files]

    for filename in files:
        # Emulate 'chmod -fR a+rX,u+w,g-w,o-w' with minimal chmod calls.
        apply_recursive_permissions(
            filename, filemode=0o644, filemask=0o022, dirmode=0o755, dirmask=0o022
        )

    return os.EX_OK


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright 2012-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class TargetRootTestCase(TestCase):
    def testTargetRoot(self):
        ebuilds = {
            "dev-lang/python-3.2": {
                "EAPI": "7",
                "BDEPEND": "~dev-lang/python-3.2",
            },
            "dev-libs/A-1": {
                "EAPI": "4",
                "DEPEND": "dev-libs/B",
                "RDEPEND": "dev-libs/C",
            },
            "dev-libs/B-1": {},
            "dev-libs/C-1": {},
        }

        installed = {
            "dev-lang/python-3.2": {
                "EAPI": "7",
                "BDEPEND": "~dev-lang/python-3.2",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-lang/python"],
                options={},
                success=True,
                mergelist=["dev-lang/python-3.2{targetroot}"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-lang/python"],
                options={"--root-deps": True},
                success=True,
                mergelist=["dev-lang/python-3.2{targetroot}"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-lang/python"],
                options={"--root-deps": "rdeps"},
                success=True,
                mergelist=["dev-lang/python-3.2{targetroot}"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={},
                ambiguous_merge_order=True,
                success=True,
                mergelist=[
                    ("dev-libs/B-1", "dev-libs/C-1{targetroot}"),
                    "dev-libs/A-1{targetroot}",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--root-deps": True},
                ambiguous_merge_order=True,
                success=True,
                mergelist=[
                    (
                        "dev-libs/B-1",
                        "dev-libs/B-1{targetroot}",
                        "dev-libs/C-1{targetroot}",
                    ),
                    "dev-libs/A-1{targetroot}",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--root-deps": "rdeps"},
                ambiguous_merge_order=True,
                success=True,
                mergelist=[("dev-libs/C-1{targetroot}"), "dev-libs/A-1{targetroot}"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, targetroot=True, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-lang/python"],
                options={},
                success=True,
                mergelist=["dev-lang/python-3.2"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, targetroot=False, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import portage


def show_lru_cache_info():
    if not os.environ.get("PORTAGE_SHOW_LRU_CACHE_INFO"):
        return

    portage_lru_caches = {
        portage.dep._use_reduce_cached: "use_reduce_cached",
        portage.eapi._get_eapi_attrs: "get_eapi_attrs",
        portage.process._encoded_length: "encoded_length",
        portage.versions.catpkgsplit: "catpkgsplit",
        portage.versions.vercmp: "vercmp",
    }

    print("Portage @lru_cache information")
    for method, name in portage_lru_caches.items():
        cache_info = method.cache_info()

        hits = cache_info.hits
        misses = cache_info.misses
        maxsize = cache_info.maxsize
        currsize = cache_info.currsize

        total = hits + misses
        if total:
            hitratio = hits / total
        else:
            hitratio = 0

        if maxsize:
            utilization = currsize / maxsize
        else:
            utilization = 0

        pretty_cache_info = f"hit ratio: {hitratio:.2%} (total: {total}, hits: {hits}, misses: {misses}) util: {utilization:.2%} ({maxsize} / {currsize})"

        print(f"{name}: {pretty_cache_info}")



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = ("VirtualsManager",)

from copy import deepcopy

from portage import os
from portage.dep import Atom
from portage.exception import InvalidAtom
from portage.localization import _
from portage.util import grabdict, stack_dictlist, writemsg
from portage.versions import cpv_getkey


class VirtualsManager:
    def __init__(self, *args, **kwargs):
        if kwargs.get("_copy"):
            return

        assert len(args) == 1, "VirtualsManager.__init__ takes one positional argument"
        assert not kwargs, (
            "unknown keyword argument(s) '%s' passed to VirtualsManager.__init__"
            % ", ".join(kwargs)
        )

        profiles = args[0]
        self._virtuals = None
        self._dirVirtuals = None
        self._virts_p = None

        # Virtuals obtained from the vartree
        self._treeVirtuals = None
        # Virtuals added by the depgraph via self.add_depgraph_virtuals().
        self._depgraphVirtuals = {}

        # Initialise _dirVirtuals.
        self._read_dirVirtuals(profiles)

        # We could initialise _treeVirtuals here, but some consumers want to
        # pass their own vartree.

    def _read_dirVirtuals(self, profiles):
        """
        Read the 'virtuals' file in all profiles.
        """
        virtuals_list = []
        for x in profiles:
            virtuals_file = os.path.join(x, "virtuals")
            virtuals_dict = grabdict(virtuals_file)
            atoms_dict = {}
            for k, v in virtuals_dict.items():
                try:
                    virt_atom = Atom(k)
                except InvalidAtom:
                    virt_atom = None
                else:
                    if virt_atom.blocker or str(virt_atom) != str(virt_atom.cp):
                        virt_atom = None
                if virt_atom is None:
                    writemsg(
                        _("--- Invalid virtuals atom in %s: %s\n") % (virtuals_file, k),
                        noiselevel=-1,
                    )
                    continue
                providers = []
                for atom in v:
                    atom_orig = atom
                    if atom[:1] == "-":
                        # allow incrementals
                        atom = atom[1:]
                    try:
                        atom = Atom(atom)
                    except InvalidAtom:
                        atom = None
                    else:
                        if atom.blocker:
                            atom = None
                    if atom is None:
                        writemsg(
                            _("--- Invalid atom in %s: %s\n")
                            % (virtuals_file, atom_orig),
                            noiselevel=-1,
                        )
                    else:
                        if atom_orig == str(atom):
                            # normal atom, so return as Atom instance
                            providers.append(atom)
                        else:
                            # atom has special prefix, so return as string
                            providers.append(atom_orig)
                if providers:
                    atoms_dict[virt_atom] = providers
            if atoms_dict:
                virtuals_list.append(atoms_dict)

        self._dirVirtuals = stack_dictlist(virtuals_list, incremental=True)

        for virt in self._dirVirtuals:
            # Preference for virtuals decreases from left to right.
            self._dirVirtuals[virt].reverse()

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}
        result = VirtualsManager(_copy=True)
        memo[id(self)] = result

        # immutable attributes (internal policy ensures lack of mutation)
        # _treeVirtuals is initialised by _populate_treeVirtuals().
        # Before that it's 'None'.
        result._treeVirtuals = self._treeVirtuals
        memo[id(self._treeVirtuals)] = self._treeVirtuals
        # _dirVirtuals is initialised by __init__.
        result._dirVirtuals = self._dirVirtuals
        memo[id(self._dirVirtuals)] = self._dirVirtuals

        # mutable attributes (change when add_depgraph_virtuals() is called)
        result._virtuals = deepcopy(self._virtuals, memo)
        result._depgraphVirtuals = deepcopy(self._depgraphVirtuals, memo)
        result._virts_p = deepcopy(self._virts_p, memo)

        return result

    def _compile_virtuals(self):
        """Stack installed and profile virtuals.  Preference for virtuals
        decreases from left to right.
        Order of preference:
        1. installed and in profile
        2. installed only
        3. profile only
        """

        assert self._treeVirtuals is not None, (
            "_populate_treeVirtuals() must be called before "
            + "any query about virtuals"
        )

        # Virtuals by profile+tree preferences.
        ptVirtuals = {}

        for virt, installed_list in self._treeVirtuals.items():
            profile_list = self._dirVirtuals.get(virt, None)
            if not profile_list:
                continue
            for cp in installed_list:
                if cp in profile_list:
                    ptVirtuals.setdefault(virt, [])
                    ptVirtuals[virt].append(cp)

        virtuals = stack_dictlist(
            [ptVirtuals, self._treeVirtuals, self._dirVirtuals, self._depgraphVirtuals]
        )
        self._virtuals = virtuals
        self._virts_p = None

    def getvirtuals(self):
        """
        Computes self._virtuals if necessary and returns it.
        self._virtuals is only computed on the first call.
        """
        if self._virtuals is None:
            self._compile_virtuals()

        return self._virtuals

    def _populate_treeVirtuals(self, vartree):
        """
        Initialize _treeVirtuals from the given vartree.
        It must not have been initialized already, otherwise
        our assumptions about immutability don't hold.
        """
        assert self._treeVirtuals is None, "treeVirtuals must not be reinitialized"

        self._treeVirtuals = {}

        for provide, cpv_list in vartree.get_all_provides().items():
            try:
                provide = Atom(provide)
            except InvalidAtom:
                continue
            self._treeVirtuals[provide.cp] = [Atom(cpv_getkey(cpv)) for cpv in cpv_list]

    def populate_treeVirtuals_if_needed(self, vartree):
        """
        Initialize _treeVirtuals if it hasn't been done already.
        This is a hack for consumers that already have an populated vartree.
        """
        if self._treeVirtuals is not None:
            return

        self._populate_treeVirtuals(vartree)

    def add_depgraph_virtuals(self, mycpv, virts):
        """This updates the preferences for old-style virtuals,
        affecting the behavior of dep_expand() and dep_check()
        calls. It can change dbapi.match() behavior since that
        calls dep_expand(). However, dbapi instances have
        internal match caches that are not invalidated when
        preferences are updated here. This can potentially
        lead to some inconsistency (relevant to bug #1343)."""

        # Ensure that self._virtuals is populated.
        if self._virtuals is None:
            self.getvirtuals()

        modified = False
        cp = Atom(cpv_getkey(mycpv))
        for virt in virts:
            try:
                virt = Atom(virt).cp
            except InvalidAtom:
                continue
            providers = self._virtuals.get(virt)
            if providers and cp in providers:
                continue
            providers = self._depgraphVirtuals.get(virt)
            if providers is None:
                providers = []
                self._depgraphVirtuals[virt] = providers
            if cp not in providers:
                providers.append(cp)
                modified = True

        if modified:
            self._compile_virtuals()

    def get_virts_p(self):
        if self._virts_p is not None:
            return self._virts_p

        virts = self.getvirtuals()
        virts_p = {}
        for x in virts:
            vkeysplit = x.split("/")
            if vkeysplit[1] not in virts_p:
                virts_p[vkeysplit[1]] = virts[x]
        self._virts_p = virts_p
        return virts_p



# Copyright 2010-2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class BacktrackingTestCase(TestCase):
    def testBacktracking(self):
        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {},
            "dev-libs/B-1": {"DEPEND": "dev-libs/A"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1", "dev-libs/B"],
                all_permutations=True,
                mergelist=["dev-libs/A-1", "dev-libs/B-1"],
                success=True,
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBacktrackNotNeeded(self):
        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {},
            "dev-libs/B-1": {},
            "dev-libs/B-2": {},
            "dev-libs/C-1": {"DEPEND": "dev-libs/A dev-libs/B"},
            "dev-libs/D-1": {"DEPEND": "=dev-libs/A-1 =dev-libs/B-1"},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/C", "dev-libs/D"],
                all_permutations=True,
                options={"--backtrack": 1},
                mergelist=[
                    "dev-libs/A-1",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBacktrackWithoutUpdates(self):
        """
        If --update is not given we might have to mask the old installed version later.
        """

        ebuilds = {
            "dev-libs/A-1": {"DEPEND": "dev-libs/Z"},
            "dev-libs/B-1": {"DEPEND": ">=dev-libs/Z-2"},
            "dev-libs/Z-1": {},
            "dev-libs/Z-2": {},
        }

        installed = {
            "dev-libs/Z-1": {"USE": ""},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/B", "dev-libs/A"],
                all_permutations=True,
                mergelist=[
                    "dev-libs/Z-2",
                    "dev-libs/B-1",
                    "dev-libs/A-1",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBacktrackMissedUpdates(self):
        """
        An update is missed due to a dependency on an older version.
        """

        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {},
            "dev-libs/B-1": {"RDEPEND": "<=dev-libs/A-1"},
        }

        installed = {
            "dev-libs/A-1": {"USE": ""},
            "dev-libs/B-1": {"USE": "", "RDEPEND": "<=dev-libs/A-1"},
        }

        options = {"--update": True, "--deep": True, "--selective": True}

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A", "dev-libs/B"],
                options=options,
                all_permutations=True,
                mergelist=[],
                success=True,
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBacktrackNoWrongRebuilds(self):
        """
        Ensure we remove backtrack masks if the reason for the mask gets masked itself.
        """

        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/D"},
            "dev-libs/C-1": {},
            "dev-libs/C-2": {"RDEPEND": ">=dev-libs/A-2"},
            "dev-libs/D-1": {"RDEPEND": "<dev-libs/A-2"},
        }

        installed = {
            "dev-libs/A-1": {},
            "dev-libs/B-1": {"RDEPEND": "dev-libs/D"},
            "dev-libs/C-1": {},
            "dev-libs/D-1": {"RDEPEND": "<dev-libs/A-2"},
        }

        world = ["dev-libs/B", "dev-libs/C"]

        options = {
            "--backtrack": 6,
            "--deep": True,
            "--selective": True,
            "--update": True,
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"], options=options, mergelist=[], success=True
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.util.endian.decode import (
    decode_uint16_le,
    decode_uint32_le,
    decode_uint16_be,
    decode_uint32_be,
)
from portage.util.elf.constants import (
    E_ENTRY,
    E_MACHINE,
    E_TYPE,
    EI_CLASS,
    ELFCLASS32,
    ELFCLASS64,
    ELFDATA2LSB,
    ELFDATA2MSB,
)


class ELFHeader:
    __slots__ = ("e_flags", "e_machine", "e_type", "ei_class", "ei_data")

    @classmethod
    def read(cls, f):
        """
        @param f: an open ELF file
        @type f: file
        @rtype: ELFHeader
        @return: A new ELFHeader instance containing data from f
        """
        f.seek(EI_CLASS)
        ei_class = ord(f.read(1))
        ei_data = ord(f.read(1))

        if ei_class == ELFCLASS32:
            width = 32
        elif ei_class == ELFCLASS64:
            width = 64
        else:
            width = None

        if ei_data == ELFDATA2LSB:
            uint16 = decode_uint16_le
            uint32 = decode_uint32_le
        elif ei_data == ELFDATA2MSB:
            uint16 = decode_uint16_be
            uint32 = decode_uint32_be
        else:
            uint16 = None
            uint32 = None

        if width is None or uint16 is None:
            e_flags = None
            e_machine = None
            e_type = None
        else:
            f.seek(E_TYPE)
            e_type = uint16(f.read(2))
            f.seek(E_MACHINE)
            e_machine = uint16(f.read(2))

            # E_ENTRY + 3 * sizeof(uintN)
            e_flags_offset = E_ENTRY + 3 * width // 8
            f.seek(e_flags_offset)
            e_flags = uint32(f.read(4))

        obj = cls()
        obj.e_flags = e_flags
        obj.e_machine = e_machine
        obj.e_type = e_type
        obj.ei_class = ei_class
        obj.ei_data = ei_data

        return obj



# Copyright 2022-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class VariableSetTestCase(TestCase):
    def testVariableSetEmerge(self):
        rust_with_rustc_rebuild = "{class=portage.sets.dbapi.VariableSet,variable=BDEPEND,includes=dev-lang/rust dev-lang/rust-bin}"

        ebuilds = {
            "dev-go/go-pkg-1": {"BDEPEND": "dev-lang/go"},
            "www-client/firefox-1": {
                "BDEPEND": "|| ( dev-lang/rust dev-lang/rust-bin )"
            },
            "dev-lang/rust-1": {"BDEPEND": "|| ( dev-lang/rust dev-lang/rust-bin )"},
            "dev-lang/rust-bin-1": {
                "BDEPEND": "|| ( dev-lang/rust-bin dev-lang/rust )"
            },
        }
        installed = ebuilds
        playground = ResolverPlayground(ebuilds=ebuilds, installed=installed)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@golang-rebuild"],
                mergelist=["dev-go/go-pkg-1"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["@rust-rebuild"],
                mergelist=["www-client/firefox-1"],
                success=True,
            ),
            ResolverPlaygroundTestCase(
                [f"@rust-with-rustc-rebuild{rust_with_rustc_rebuild}"],
                mergelist=[
                    "www-client/firefox-1",
                    "dev-lang/rust-1",
                    "dev-lang/rust-bin-1",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
        )

        try:
            for test_case in test_cases:
                # Create an artificial VariableSet to test against
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2024  Alexey Gladkov <gladkov.alexey@gmail.com>

doc = """Zipfile plug-in module for portage.
Performs a http download of a portage snapshot and unpacks it to the repo
location."""
__doc__ = doc[:]


import os

from portage.sync.config_checks import CheckSyncConfig


module_spec = {
    "name": "zipfile",
    "description": doc,
    "provides": {
        "zipfile-module": {
            "name": "zipfile",
            "sourcefile": "zipfile",
            "class": "ZipFile",
            "description": doc,
            "functions": ["sync", "retrieve_head"],
            "func_desc": {
                "sync": "Performs an archived http download of the "
                + "repository, then unpacks it.",
                "retrieve_head": "Returns the checksum of the unpacked archive.",
            },
            "validate_config": CheckSyncConfig,
            "module_specific_options": (),
        },
    },
}



# Copyright 2010-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import collections
from functools import lru_cache
from typing import Optional

from portage import eapi_is_supported


def eapi_has_iuse_defaults(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).iuse_defaults


def eapi_has_iuse_effective(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).iuse_effective


def eapi_has_slot_deps(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).slot_deps


def eapi_has_slot_operator(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).slot_operator


def eapi_has_src_uri_arrows(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).src_uri_arrows


def eapi_has_selective_src_uri_restriction(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).selective_src_uri_restriction


def eapi_has_use_deps(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).use_deps


def eapi_has_strong_blocks(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).strong_blocks


def eapi_has_src_prepare_and_src_configure(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).src_prepare_src_configure


def eapi_supports_prefix(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).prefix


def eapi_exports_pms_vars(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_pms_vars


def eapi_exports_AA(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_AA


def eapi_exports_KV(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_KV


def eapi_exports_merge_type(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_merge_type


def eapi_exports_replace_vars(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_replace_vars


def eapi_exports_EBUILD_PHASE_FUNC(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_EBUILD_PHASE_FUNC


def eapi_exports_PORTDIR(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_PORTDIR


def eapi_exports_ECLASSDIR(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).exports_ECLASSDIR


def eapi_has_pkg_pretend(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).pkg_pretend


def eapi_has_implicit_rdepend(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).rdepend_depend


def eapi_has_dosed_dohard(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).dosed_dohard


def eapi_has_required_use(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).required_use


def eapi_has_required_use_at_most_one_of(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).required_use_at_most_one_of


def eapi_has_use_dep_defaults(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).use_dep_defaults


def eapi_requires_posixish_locale(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).posixish_locale


def eapi_has_repo_deps(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).repo_deps


def eapi_supports_stable_use_forcing_and_masking(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).stablemask


def eapi_allows_directories_on_profile_level_and_repository_level(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).profile_file_dirs


def eapi_allows_package_provided(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).allows_package_provided


def eapi_has_bdepend(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).bdepend


def eapi_has_idepend(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).idepend


def eapi_empty_groups_always_true(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).empty_groups_always_true


def eapi_path_variables_end_with_trailing_slash(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).path_variables_end_with_trailing_slash


def eapi_has_broot(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).broot


def eapi_has_sysroot(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).sysroot


def eapi_rewrites_symlinks(eapi: str) -> bool:
    return _get_eapi_attrs(eapi).symlink_rewrite


_eapi_attrs = collections.namedtuple(
    "_eapi_attrs",
    (
        "allows_package_provided",
        "bdepend",
        "broot",
        "dosed_dohard",
        "empty_groups_always_true",
        "exports_AA",
        "exports_EBUILD_PHASE_FUNC",
        "exports_ECLASSDIR",
        "exports_KV",
        "exports_merge_type",
        "exports_pms_vars",
        "exports_PORTDIR",
        "exports_replace_vars",
        "feature_flag_test",
        "idepend",
        "iuse_defaults",
        "iuse_effective",
        "posixish_locale",
        "path_variables_end_with_trailing_slash",
        "pkg_pretend",
        "prefix",
        "profile_file_dirs",
        "rdepend_depend",
        "repo_deps",
        "required_use",
        "required_use_at_most_one_of",
        "selective_src_uri_restriction",
        "slot_operator",
        "slot_deps",
        "src_prepare_src_configure",
        "src_uri_arrows",
        "stablemask",
        "strong_blocks",
        "symlink_rewrite",
        "sysroot",
        "use_deps",
        "use_dep_defaults",
    ),
)


class Eapi:
    ALL_EAPIS = (
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
    )

    _eapi_val: int = -1

    def __init__(self, eapi_string: str):
        if not eapi_string in self.ALL_EAPIS:
            raise ValueError(f"'{eapi_string}' not recognized as a valid EAPI")

        self._eapi_val = int(eapi_string.partition("-")[0])

    def __ge__(self, other: "Eapi") -> bool:
        return self._eapi_val >= other._eapi_val

    def __le__(self, other: "Eapi") -> bool:
        return self._eapi_val <= other._eapi_val


@lru_cache(32)
def _get_eapi_attrs(eapi_str: Optional[str]) -> _eapi_attrs:
    """
    When eapi is None then validation is not as strict, since we want the
    same to work for multiple EAPIs that may have slightly different rules.
    An unsupported eapi is handled the same as when eapi is None, which may
    be helpful for handling of corrupt EAPI metadata in essential functions
    such as pkgsplit.
    """
    if eapi_str is None or not eapi_is_supported(eapi_str):
        return _eapi_attrs(
            allows_package_provided=True,
            bdepend=False,
            broot=True,
            dosed_dohard=False,
            empty_groups_always_true=False,
            exports_AA=False,
            exports_EBUILD_PHASE_FUNC=True,
            exports_ECLASSDIR=False,
            exports_KV=False,
            exports_merge_type=True,
            exports_pms_vars=True,
            exports_PORTDIR=True,
            exports_replace_vars=True,
            feature_flag_test=False,
            idepend=False,
            iuse_defaults=True,
            iuse_effective=False,
            path_variables_end_with_trailing_slash=False,
            pkg_pretend=True,
            posixish_locale=False,
            prefix=True,
            profile_file_dirs=False,
            rdepend_depend=False,
            repo_deps=True,
            required_use=True,
            required_use_at_most_one_of=True,
            selective_src_uri_restriction=True,
            slot_deps=True,
            slot_operator=True,
            src_prepare_src_configure=True,
            src_uri_arrows=True,
            stablemask=True,
            strong_blocks=True,
            symlink_rewrite=False,
            sysroot=True,
            use_deps=True,
            use_dep_defaults=True,
        )
    else:
        eapi = Eapi(eapi_str)
        return _eapi_attrs(
            allows_package_provided=eapi <= Eapi("6"),
            bdepend=eapi >= Eapi("7"),
            broot=eapi >= Eapi("7"),
            dosed_dohard=eapi <= Eapi("3"),
            empty_groups_always_true=eapi <= Eapi("6"),
            exports_AA=eapi <= Eapi("3"),
            exports_EBUILD_PHASE_FUNC=eapi >= Eapi("5"),
            exports_ECLASSDIR=eapi <= Eapi("6"),
            exports_KV=eapi <= Eapi("3"),
            exports_merge_type=eapi >= Eapi("4"),
            exports_pms_vars=eapi <= Eapi("8"),
            exports_PORTDIR=eapi <= Eapi("6"),
            exports_replace_vars=eapi >= Eapi("4"),
            feature_flag_test=False,
            idepend=eapi >= Eapi("8"),
            iuse_defaults=eapi >= Eapi("1"),
            iuse_effective=eapi >= Eapi("5"),
            path_variables_end_with_trailing_slash=eapi <= Eapi("6"),
            pkg_pretend=eapi >= Eapi("4"),
            posixish_locale=eapi >= Eapi("6"),
            prefix=eapi >= Eapi("3"),
            profile_file_dirs=eapi >= Eapi("7"),
            rdepend_depend=eapi <= Eapi("3"),
            repo_deps=False,
            required_use=eapi >= Eapi("4"),
            required_use_at_most_one_of=eapi >= Eapi("5"),
            selective_src_uri_restriction=eapi >= Eapi("8"),
            slot_deps=eapi >= Eapi("1"),
            slot_operator=eapi >= Eapi("5"),
            src_prepare_src_configure=eapi >= Eapi("2"),
            src_uri_arrows=eapi >= Eapi("2"),
            stablemask=eapi >= Eapi("5"),
            strong_blocks=eapi >= Eapi("2"),
            symlink_rewrite=eapi <= Eapi("8"),
            sysroot=eapi >= Eapi("7"),
            use_deps=eapi >= Eapi("2"),
            use_dep_defaults=eapi >= Eapi("4"),
        )



# Copyright 1999-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import logging
import shlex

import portage
from portage import os
from portage.util import writemsg_level
from portage.sync.syncbase import NewBase


class CVSSync(NewBase):
    """CVS sync module"""

    short_desc = "Perform sync operations on CVS repositories"

    @staticmethod
    def name():
        return "CVSSync"

    def __init__(self):
        NewBase.__init__(self, "cvs", portage.const.CVS_PACKAGE_ATOM)

    def exists(self, **kwargs):
        """Tests whether the repo is checked out"""
        return os.path.exists(os.path.join(self.repo.location, "CVS"))

    def new(self, **kwargs):
        if kwargs:
            self._kwargs(kwargs)
        # initial checkout
        cvs_root = self.repo.sync_uri
        if (
            portage.process.spawn_bash(
                "cd %s; exec cvs -z0 -d %s co -P -d %s %s"
                % (
                    shlex.quote(os.path.dirname(self.repo.location)),
                    shlex.quote(cvs_root),
                    shlex.quote(os.path.basename(self.repo.location)),
                    shlex.quote(self.repo.module_specific_options["sync-cvs-repo"]),
                ),
                **self.spawn_kwargs,
            )
            != os.EX_OK
        ):
            msg = "!!! cvs checkout error; exiting."
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", noiselevel=-1, level=logging.ERROR)
            return (1, False)
        return (0, False)

    def update(self):
        """
        Internal function to update an existing CVS repository

        @return: tuple of return code (0=success), whether the cache
                needs to be updated
        @rtype: (int, bool)
        """

        # cvs update
        exitcode = portage.process.spawn_bash(
            f"cd {shlex.quote(self.repo.location)}; exec cvs -z0 -q update -dP",
            **self.spawn_kwargs,
        )
        if exitcode != os.EX_OK:
            msg = "!!! cvs update error; exiting."
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n", noiselevel=-1, level=logging.ERROR)
        return (exitcode, False)



# Copyright 1999-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import io
import sys

from _emerge.CompositeTask import CompositeTask
from _emerge.EbuildPhase import EbuildPhase

import portage
from portage import os


class EbuildBinpkg(CompositeTask):
    """
    This assumes that src_install() has successfully completed.
    """

    __slots__ = ("pkg", "settings") + (
        "_binpkg_tmpfile",
        "_binpkg_info",
        "pkg_allocated_path",
    )

    def _start(self):
        pkg = self.pkg
        root_config = pkg.root_config
        bintree = root_config.trees["bintree"]
        pkg_allocated_path, build_id = bintree.getname_build_id(
            pkg.cpv, allocate_new=True
        )
        bintree._ensure_dir(os.path.dirname(pkg_allocated_path))

        self.pkg_allocated_path = pkg_allocated_path
        self._binpkg_tmpfile = self.pkg_allocated_path + "." + str(portage.getpid())
        self.settings["PORTAGE_BINPKG_TMPFILE"] = self._binpkg_tmpfile

        if "binpkg-multi-instance" in self.settings.features:
            self.settings["BUILD_ID"] = str(build_id)

        package_phase = EbuildPhase(
            background=self.background,
            phase="package",
            scheduler=self.scheduler,
            settings=self.settings,
        )

        self._start_task(package_phase, self._package_phase_exit)

    def _package_phase_exit(self, package_phase):
        self.settings.pop("PORTAGE_BINPKG_TMPFILE", None)
        if self._default_exit(package_phase) != os.EX_OK:
            try:
                os.unlink(self._binpkg_tmpfile)
            except OSError:
                pass
            self.wait()
            return

        pkg = self.pkg
        bintree = pkg.root_config.trees["bintree"]

        stdout_orig = sys.stdout
        stderr_orig = sys.stderr
        out = io.StringIO()
        try:
            sys.stdout = out
            sys.stderr = out

            self._binpkg_info = bintree.inject(
                pkg.cpv,
                current_pkg_path=self._binpkg_tmpfile,
                allocated_pkg_path=self.pkg_allocated_path,
            )

        finally:
            sys.stdout = stdout_orig
            sys.stderr = stderr_orig

            output_value = out.getvalue()
            if output_value:
                self.scheduler.output(
                    output_value,
                    log_path=self.settings.get("PORTAGE_LOG_FILE"),
                    background=self.background,
                )

        self._current_task = None
        self.returncode = 1 if self._binpkg_info is None else os.EX_OK
        self.wait()

    def get_binpkg_info(self):
        return self._binpkg_info



# Copyright 2020-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import multiprocessing
import sys

import portage
from portage import os
from portage.tests import TestCase
from portage.util._async.AsyncFunction import AsyncFunction
from portage.util.futures import asyncio
from portage.util.futures._asyncio.streams import _writer
from portage.util.futures.unix_events import _set_nonblocking


class AsyncFunctionTestCase(TestCase):
    @staticmethod
    def _read_from_stdin(pw):
        if pw is not None:
            os.close(pw)
        return "".join(sys.stdin)

    async def _testAsyncFunctionStdin(self, loop):
        test_string = "1\n2\n3\n"
        pr, pw = multiprocessing.Pipe(duplex=False)
        stdin_backup = os.dup(portage._get_stdin().fileno())
        os.dup2(pr.fileno(), portage._get_stdin().fileno())
        pr.close()
        try:
            reader = AsyncFunction(
                # Should automatically inherit stdin as fd_pipes[0]
                # when background is False, for things like
                # emerge --sync --ask (bug 916116).
                background=False,
                scheduler=loop,
                target=self._read_from_stdin,
                args=(
                    (
                        pw.fileno()
                        if multiprocessing.get_start_method() == "fork"
                        else None
                    ),
                ),
            )
            reader.start()
        finally:
            os.dup2(stdin_backup, portage._get_stdin().fileno())
            os.close(stdin_backup)

        _set_nonblocking(pw.fileno())
        with open(pw.fileno(), mode="wb", buffering=0, closefd=False) as pipe_write:
            await _writer(pipe_write, test_string.encode("utf_8"))
        pw.close()
        self.assertEqual((await reader.async_wait()), os.EX_OK)
        self.assertEqual(reader.result, test_string)

    def testAsyncFunctionStdin(self):
        loop = asyncio._wrap_loop()
        loop.run_until_complete(self._testAsyncFunctionStdin(loop=loop))

    def testAsyncFunctionStdinSpawn(self):
        orig_start_method = multiprocessing.get_start_method()
        if orig_start_method == "spawn":
            self.skipTest("multiprocessing start method is already spawn")
        # NOTE: An attempt was made to use multiprocessing.get_context("spawn")
        # here, but it caused the python process to terminate unexpectedly
        # during a send_handle call.
        multiprocessing.set_start_method("spawn", force=True)
        try:
            self.testAsyncFunctionStdin()
        finally:
            multiprocessing.set_start_method(orig_start_method, force=True)

    @staticmethod
    def _test_getpid_fork(preexec_fn=None):
        """
        Verify that portage.getpid() cache is updated in a forked child process.
        """
        if preexec_fn is not None:
            preexec_fn()
        loop = asyncio._wrap_loop()
        proc = AsyncFunction(scheduler=loop, target=portage.getpid)
        proc.start()
        proc.wait()
        return proc.pid == proc.result

    def test_getpid_fork(self):
        self.assertTrue(self._test_getpid_fork())

    def test_spawn_getpid(self):
        """
        Test portage.getpid() with multiprocessing spawn start method.
        """
        loop = asyncio._wrap_loop()
        proc = AsyncFunction(
            scheduler=loop,
            target=self._test_getpid_fork,
            kwargs=dict(
                preexec_fn=functools.partial(
                    multiprocessing.set_start_method, "spawn", force=True
                )
            ),
        )
        proc.start()
        self.assertEqual(proc.wait(), 0)
        self.assertTrue(proc.result)

    def test_getpid_double_fork(self):
        """
        Verify that portage.getpid() cache is updated correctly after
        two forks.
        """
        loop = asyncio._wrap_loop()
        proc = AsyncFunction(scheduler=loop, target=self._test_getpid_fork)
        proc.start()
        self.assertEqual(proc.wait(), 0)
        self.assertTrue(proc.result)



# Copyright 2018-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ("ForkExecutor",)

import collections
import functools
import os
import traceback

from portage.util._async.AsyncFunction import AsyncFunction
from portage.util.futures import asyncio
from portage.util.cpuinfo import get_cpu_count


class ForkExecutor:
    """
    An implementation of concurrent.futures.Executor that forks a
    new process for each task, with support for cancellation of tasks.

    This is entirely driven by an event loop.
    """

    def __init__(self, max_workers=None, loop=None):
        self._max_workers = max_workers or get_cpu_count()
        self._loop = asyncio._wrap_loop(loop)
        self._submit_queue = collections.deque()
        self._running_tasks = {}
        self._shutdown = False
        self._shutdown_future = self._loop.create_future()

    def submit(self, fn, *args, **kwargs):
        """Submits a callable to be executed with the given arguments.

        Schedules the callable to be executed as fn(*args, **kwargs) and returns
        a Future instance representing the execution of the callable.

        Returns:
                A Future representing the given call.
        """
        future = self._loop.create_future()
        proc = AsyncFunction(
            target=functools.partial(self._guarded_fn_call, fn, args, kwargs),
            # Directly inherit stdio streams and run in the foreground with no log.
            create_pipe=False,
        )
        self._submit_queue.append((future, proc))
        self._schedule()
        return future

    def _schedule(self):
        while (
            not self._shutdown
            and self._submit_queue
            and (
                self._max_workers is True
                or len(self._running_tasks) < self._max_workers
            )
        ):
            future, proc = self._submit_queue.popleft()
            future.add_done_callback(functools.partial(self._cancel_cb, proc))
            proc.addExitListener(functools.partial(self._proc_exit, future))
            proc.scheduler = self._loop
            proc.start()
            self._running_tasks[id(proc)] = proc

    def _cancel_cb(self, proc, future):
        if future.cancelled():
            # async, handle the rest in _proc_exit
            proc.cancel()

    @staticmethod
    def _guarded_fn_call(fn, args, kwargs):
        try:
            result = fn(*args, **kwargs)
            exception = None
        except Exception as e:
            result = None
            exception = _ExceptionWithTraceback(e)

        return result, exception

    def _proc_exit(self, future, proc):
        if not future.cancelled():
            if proc.returncode == os.EX_OK:
                result, exception = proc.result
                if exception is not None:
                    future.set_exception(exception)
                else:
                    future.set_result(result)
            else:
                # TODO: add special exception class for this, maybe
                # distinguish between kill and crash
                future.set_exception(
                    Exception(
                        f"pid {proc.pid} crashed or killed, exitcode {proc.returncode}"
                    )
                )

        del self._running_tasks[id(proc)]
        self._schedule()
        if self._shutdown and not self._running_tasks:
            self._shutdown_future.set_result(None)

    def shutdown(self, wait=True):
        self._shutdown = True
        if not self._running_tasks and not self._shutdown_future.done():
            self._shutdown_future.set_result(None)
        if wait:
            self._loop.run_until_complete(self._shutdown_future)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)
        return False


class _ExceptionWithTraceback:
    def __init__(self, exc):
        tb = traceback.format_exception(type(exc), exc, exc.__traceback__)
        tb = "".join(tb)
        self.exc = exc
        self.tb = f'\n"""\n{tb}"""'

    def __reduce__(self):
        return _rebuild_exc, (self.exc, self.tb)


class _RemoteTraceback(Exception):
    def __init__(self, tb):
        self.tb = tb

    def __str__(self):
        return self.tb


def _rebuild_exc(exc, tb):
    exc.__cause__ = _RemoteTraceback(tb)
    return exc



# Copyright 2010-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import io

import portage
from portage import os
from portage.dep import Atom, _repo_name_re
from portage.eapi import eapi_has_repo_deps
from portage.elog import messages as elog_messages
from portage.exception import InvalidAtom
from portage.package.ebuild._ipc.IpcCommand import IpcCommand
from portage.util import normalize_path, no_color
from portage.versions import best


class QueryCommand(IpcCommand):
    __slots__ = (
        "phase",
        "settings",
    )

    _db = None

    @classmethod
    def get_db(cls):
        if cls._db is not None:
            return cls._db
        return portage.db

    def __init__(self, settings, phase):
        IpcCommand.__init__(self)
        self.settings = settings
        self.phase = phase

    def __call__(self, argv):
        """
        @return: tuple of (stdout, stderr, returncode)
        """

        # Python 3:
        # cmd, root, *args = argv
        cmd = argv[0]
        root = argv[1]
        args = argv[2:]

        warnings = []
        warnings_str = ""

        db = self.get_db()
        eapi = self.settings.get("EAPI")

        root = normalize_path(root or os.sep).rstrip(os.sep) + os.sep
        if root not in db:
            return ("", f"{cmd}: Invalid ROOT: {root}\n", 3)

        portdb = db[root]["porttree"].dbapi
        vardb = db[root]["vartree"].dbapi

        if cmd in ("best_version", "has_version"):
            allow_repo = eapi_has_repo_deps(eapi)
            try:
                atom = Atom(args[0], allow_repo=allow_repo)
            except InvalidAtom:
                return ("", f"{cmd}: Invalid atom: {args[0]}\n", 2)

            try:
                atom = Atom(args[0], allow_repo=allow_repo, eapi=eapi)
            except InvalidAtom as e:
                warnings.append(f"QA Notice: {cmd}: {e}")

            use = self.settings.get("PORTAGE_BUILT_USE")
            if use is None:
                use = self.settings["PORTAGE_USE"]

            use = frozenset(use.split())
            atom = atom.evaluate_conditionals(use)

        if warnings:
            warnings_str = self._elog("eqawarn", warnings)

        if cmd == "has_version":
            if vardb.match(atom):
                returncode = 0
            else:
                returncode = 1
            return ("", warnings_str, returncode)
        if cmd == "best_version":
            m = best(vardb.match(atom))
            return (f"{m}\n", warnings_str, 0)
        if cmd in (
            "master_repositories",
            "repository_path",
            "available_eclasses",
            "eclass_path",
            "license_path",
        ):
            repo = _repo_name_re.match(args[0])
            if repo is None:
                return ("", f"{cmd}: Invalid repository: {args[0]}\n", 2)
            try:
                repo = portdb.repositories[args[0]]
            except KeyError:
                return ("", warnings_str, 1)

            if cmd == "master_repositories":
                return (
                    f"{' '.join(x.name for x in repo.masters)}\n",
                    warnings_str,
                    0,
                )
            if cmd == "repository_path":
                return (f"{repo.location}\n", warnings_str, 0)
            if cmd == "available_eclasses":
                return (
                    f"{' '.join(sorted(repo.eclass_db.eclasses))}\n",
                    warnings_str,
                    0,
                )
            if cmd == "eclass_path":
                try:
                    eclass = repo.eclass_db.eclasses[args[1]]
                except KeyError:
                    return ("", warnings_str, 1)
                return (f"{eclass.location}\n", warnings_str, 0)
            if cmd == "license_path":
                paths = reversed(
                    [
                        os.path.join(x.location, "licenses", args[1])
                        for x in list(repo.masters) + [repo]
                    ]
                )
                for path in paths:
                    if os.path.exists(path):
                        return (f"{path}\n", warnings_str, 0)
                return ("", warnings_str, 1)
        return ("", f"Invalid command: {cmd}\n", 3)

    def _elog(self, elog_funcname, lines):
        """
        This returns a string, to be returned via ipc and displayed at the
        appropriate place in the build output. We wouldn't want to open the
        log here since it is already opened by AbstractEbuildProcess and we
        don't want to corrupt it, especially if it is being written with
        compression.
        """
        out = io.StringIO()
        phase = self.phase
        elog_func = getattr(elog_messages, elog_funcname)
        global_havecolor = portage.output.havecolor
        try:
            portage.output.havecolor = not no_color(self.settings)
            for line in lines:
                elog_func(line, phase=phase, key=self.settings.mycpv, out=out)
        finally:
            portage.output.havecolor = global_havecolor
        msg = out.getvalue()
        return msg



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""Tests for the portage.util._xattr module"""

from unittest import mock

import subprocess

import portage
from portage.tests import TestCase
from portage.util._xattr import xattr as _xattr, _XattrSystemCommands, _XattrStub


orig_popen = subprocess.Popen


def MockSubprocessPopen(stdin):
    """Helper to mock (closely) a subprocess.Popen call

    The module has minor tweaks in behavior when it comes to encoding and
    python versions, so use a real subprocess.Popen call to fake out the
    runtime behavior.  This way we don't have to also implement different
    encodings as that gets ugly real fast.
    """
    # pylint: disable=protected-access
    proc = orig_popen(["cat"], stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    proc.stdin.write(portage._unicode_encode(stdin, portage._encodings["stdio"]))
    return proc


class SystemCommandsTest(TestCase):
    """Test _XattrSystemCommands"""

    OUTPUT = "\n".join(
        (
            "# file: /bin/ping",
            "security.capability=0sAQAAAgAgAAAAAAAAAAAAAAAAAAA=",
            'user.foo="asdf"',
            "",
        )
    )

    def _setUp(self):
        return _XattrSystemCommands

    def _testGetBasic(self):
        """Verify the get() behavior"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify basic behavior, and namespace arg works as expected.
            xattr.get("/some/file", "user.foo")
            xattr.get("/some/file", "foo", namespace="user")
            self.assertEqual(call_mock.call_args_list[0], call_mock.call_args_list[1])

            # Verify nofollow behavior.
            call_mock.reset()
            xattr.get("/some/file", "user.foo", nofollow=True)
            self.assertIn("-h", call_mock.call_args[0][0])

    def testGetParsing(self):
        """Verify get() parses output sanely"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify output parsing.
            call_mock.return_value = MockSubprocessPopen(
                "\n".join(
                    [
                        "# file: /some/file",
                        'user.foo="asdf"',
                        "",
                    ]
                )
            )
            call_mock.reset()
            self.assertEqual(xattr.get("/some/file", "user.foo"), b'"asdf"')

    def testGetAllBasic(self):
        """Verify the get_all() behavior"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify basic behavior.
            xattr.get_all("/some/file")

            # Verify nofollow behavior.
            call_mock.reset()
            xattr.get_all("/some/file", nofollow=True)
            self.assertIn("-h", call_mock.call_args[0][0])

    def testGetAllParsing(self):
        """Verify get_all() parses output sanely"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify output parsing.
            call_mock.return_value = MockSubprocessPopen(self.OUTPUT)
            exp = [
                (b"security.capability", b"0sAQAAAgAgAAAAAAAAAAAAAAAAAAA="),
                (b"user.foo", b'"asdf"'),
            ]
            self.assertEqual(exp, xattr.get_all("/some/file"))

    def testSetBasic(self):
        """Verify the set() behavior"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify basic behavior, and namespace arg works as expected.
            xattr.set("/some/file", "user.foo", "bar")
            xattr.set("/some/file", "foo", "bar", namespace="user")
            self.assertEqual(call_mock.call_args_list[0], call_mock.call_args_list[1])

    def testListBasic(self):
        """Verify the list() behavior"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify basic behavior.
            xattr.list("/some/file")

            # Verify nofollow behavior.
            call_mock.reset()
            xattr.list("/some/file", nofollow=True)
            self.assertIn("-h", call_mock.call_args[0][0])

    def testListParsing(self):
        """Verify list() parses output sanely"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify output parsing.
            call_mock.return_value = MockSubprocessPopen(self.OUTPUT)
            exp = [b"security.capability", b"user.foo"]
            self.assertEqual(exp, xattr.list("/some/file"))

    def testRemoveBasic(self):
        """Verify the remove() behavior"""
        xattr = self._setUp()
        with mock.patch.object(subprocess, "Popen") as call_mock:
            # Verify basic behavior, and namespace arg works as expected.
            xattr.remove("/some/file", "user.foo")
            xattr.remove("/some/file", "foo", namespace="user")
            self.assertEqual(call_mock.call_args_list[0], call_mock.call_args_list[1])

            # Verify nofollow behavior.
            call_mock.reset()
            xattr.remove("/some/file", "user.foo", nofollow=True)
            self.assertIn("-h", call_mock.call_args[0][0])


class StubTest(TestCase):
    """Test _XattrStub"""

    def testBasic(self):
        """Verify the stub is stubby"""
        # Would be nice to verify raised errno is OperationNotSupported.
        self.assertRaises(OSError, _XattrStub.get, "/", "")
        self.assertRaises(OSError, _XattrStub.set, "/", "", "")
        self.assertRaises(OSError, _XattrStub.get_all, "/")
        self.assertRaises(OSError, _XattrStub.remove, "/", "")
        self.assertRaises(OSError, _XattrStub.list, "/")


class StandardTest(TestCase):
    """Test basic xattr API"""

    MODULES = (_xattr, _XattrSystemCommands, _XattrStub)
    FUNCS = ("get", "get_all", "set", "remove", "list")

    def testApi(self):
        """Make sure the exported API matches"""
        for mod in self.MODULES:
            for f in self.FUNCS:
                self.assertTrue(hasattr(mod, f), f"{f} func missing in {mod}")



# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import json
import os
from typing import Optional

import portage
from portage.locks import lockfile, unlockfile
from portage.repository.config import RepoConfig
from portage.util.path import first_existing

_HISTORY_LIMIT = 25


def get_repo_revision_history(
    eroot: str, repos: Optional[list[RepoConfig]] = None
) -> dict[str, list[str]]:
    """
    Get revision history of synced repos. Returns a dict that maps
    a repo name to list of revisions in descending order by time.
    If a change is detected and the current process has permission
    to update the repo_revisions file, then the file will be updated
    with any newly detected revisions.

    This functions detects revisions which are not yet visible to the
    current process due to the sync-rcu option.

    @param eroot: EROOT to query
    @type eroot: string
    @param repos: list of RepoConfig instances to check for new revisions
    @type repos: list
    @rtype: dict
    @return: mapping of repo name to list of revisions in descending
             order by time
    """
    items = []
    for repo in repos or ():
        if repo.volatile:
            items.append((repo, None))
            continue
        if repo.sync_type:
            try:
                sync_mod = portage.sync.module_controller.get_class(repo.sync_type)
            except portage.exception.PortageException:
                continue
        else:
            continue
        repo_location_orig = repo.location
        try:
            if repo.user_location is not None:
                # Temporarily override sync-rcu behavior which pins the
                # location to a previous snapshot, since we want the
                # latest available revision here.
                repo.location = repo.user_location
            status, repo_revision = sync_mod().retrieve_head(options={"repo": repo})
        except NotImplementedError:
            repo_revision = None
        else:
            repo_revision = repo_revision.strip() if status == os.EX_OK else None
        finally:
            repo.location = repo_location_orig

        if repo_revision is not None:
            items.append((repo, repo_revision))

    return _maybe_update_revisions(eroot, items)


def _update_revisions(repo_revisions, items):
    modified = False
    for repo, repo_revision in items:
        if repo.volatile:
            # For volatile repos the revisions may be unordered,
            # which makes them unusable here where revisions are
            # intended to be ordered, so discard them.
            rev_list = repo_revisions.pop(repo.name, None)
            if rev_list:
                modified = True
            continue

        rev_list = repo_revisions.setdefault(repo.name, [])
        if not rev_list or rev_list[0] != repo_revision:
            rev_list.insert(0, repo_revision)
            del rev_list[_HISTORY_LIMIT:]
            modified = True
    return modified


def _maybe_update_revisions(eroot, items):
    repo_revisions_file = os.path.join(eroot, portage.const.REPO_REVISIONS)
    repo_revisions_lock = None
    try:
        previous_revisions = None
        try:
            with open(repo_revisions_file, encoding="utf8") as f:
                if os.fstat(f.fileno()).st_size:
                    previous_revisions = json.load(f)
        except FileNotFoundError:
            pass

        repo_revisions = {} if previous_revisions is None else previous_revisions.copy()
        modified = _update_revisions(repo_revisions, items)

        # If modified then do over with lock if permissions allow.
        if modified and os.access(
            first_existing(os.path.dirname(repo_revisions_file)), os.W_OK
        ):
            # This is a bit redundant since the config._init_dirs method
            # is supposed to create PRIVATE_PATH with these permissions.
            portage.util.ensure_dirs(
                os.path.dirname(repo_revisions_file),
                gid=portage.data.portage_gid,
                mode=0o2750,
                mask=0o2,
            )
            repo_revisions_lock = lockfile(repo_revisions_file)
            previous_revisions = None
            with open(repo_revisions_file, encoding="utf8") as f:
                if os.fstat(f.fileno()).st_size:
                    previous_revisions = json.load(f)
            repo_revisions = (
                {} if previous_revisions is None else previous_revisions.copy()
            )
            _update_revisions(repo_revisions, items)
            f = portage.util.atomic_ofstream(repo_revisions_file)
            json.dump(repo_revisions, f, ensure_ascii=False, sort_keys=True)
            f.close()
    finally:
        if repo_revisions_lock is not None:
            unlockfile(repo_revisions_lock)

    return repo_revisions



# setup_env.py -- Make sure bin subdir has sane env for testing
# Copyright 2007-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tempfile

import portage
from portage import os
from portage import shutil
from portage.const import PORTAGE_BIN_PATH
from portage.const import PORTAGE_PYM_PATH
from portage.tests import TestCase
from portage.process import spawn

bindir = PORTAGE_BIN_PATH
basedir = None
env = None


def binTestsCleanup():
    global basedir
    if basedir is None:
        return
    if os.access(basedir, os.W_OK):
        shutil.rmtree(basedir)
        basedir = None


def binTestsInit():
    binTestsCleanup()
    global basedir, env
    basedir = tempfile.mkdtemp()
    env = {}
    env["EAPI"] = "0"
    env["D"] = os.path.join(basedir, "image")
    env["T"] = os.path.join(basedir, "temp")
    env["S"] = os.path.join(basedir, "workdir")
    env["PF"] = "portage-tests-0.09-r1"
    env["PATH"] = bindir + ":" + os.environ["PATH"]
    env["PORTAGE_BIN_PATH"] = bindir
    env["PORTAGE_PYM_PATH"] = PORTAGE_PYM_PATH
    env["PORTAGE_PYTHON"] = portage._python_interpreter
    env["PORTAGE_INST_UID"] = str(os.getuid())
    env["PORTAGE_INST_GID"] = str(os.getgid())
    env["DESTTREE"] = "/usr"
    os.mkdir(env["D"])
    os.mkdir(env["T"])
    os.mkdir(env["S"])


class BinTestCase(TestCase):
    def init(self):
        binTestsInit()

    def cleanup(self):
        binTestsCleanup()


def _exists_in_D(path):
    # Note: do not use os.path.join() here, we assume D to end in /
    return os.access(env["D"] + path, os.W_OK)


def exists_in_D(path):
    if not _exists_in_D(path):
        raise TestCase.failureException


def xexists_in_D(path):
    if _exists_in_D(path):
        raise TestCase.failureException


def portage_func(func, args, exit_status=0):
    # we don't care about the output of the programs,
    # just their exit value and the state of $D
    global env
    f = open("/dev/null", "wb")
    fd_pipes = {0: 0, 1: f.fileno(), 2: f.fileno()}

    spawn([func] + args.split(), env=env, fd_pipes=fd_pipes, cwd=env["S"])
    f.close()


def create_portage_wrapper(f):
    def derived_func(*args):
        newargs = list(args)
        newargs.insert(0, f)
        return portage_func(*newargs)

    return derived_func


for f in os.listdir(os.path.join(bindir, "ebuild-helpers")):
    if (
        f.startswith("do")
        or f.startswith("new")
        or f.startswith("prep")
        or f in ("fowners", "fperms")
    ):
        globals()[f] = create_portage_wrapper(os.path.join(bindir, "ebuild-helpers", f))



# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import signal

from portage.process import find_binary, spawn
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop


class SpawnReturnProcTestCase(TestCase):
    def testSpawnReturnProcWait(self):
        true_binary = find_binary("true")
        self.assertNotEqual(true_binary, None)

        loop = global_event_loop()

        async def watch_pid():
            proc = spawn([true_binary], returnproc=True)
            self.assertEqual(await proc.wait(), os.EX_OK)

            # A second wait should also work.
            self.assertEqual(await proc.wait(), os.EX_OK)

        loop.run_until_complete(watch_pid())

    def testSpawnReturnProcTerminate(self):
        sleep_binary = find_binary("sleep")
        self.assertNotEqual(sleep_binary, None)

        loop = global_event_loop()

        async def watch_pid():
            proc = spawn([sleep_binary, "9999"], returnproc=True)
            proc.terminate()
            self.assertEqual(await proc.wait(), -signal.SIGTERM)

        loop.run_until_complete(watch_pid())



# Copyright 1999-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import copy
import functools
import io
import multiprocessing
import signal
import sys

import portage
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage import _unicode_decode
from portage.checksum import _hash_filter
from portage.elog.messages import eerror
from portage.package.ebuild.fetch import (
    _check_distfile,
    _drop_privs_userfetch,
    _want_userfetch,
    async_fetch,
)
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
from portage.util._async.ForkProcess import ForkProcess
from portage.util._pty import _create_pty_or_pipe
from portage.util.futures import asyncio
from _emerge.CompositeTask import CompositeTask


class EbuildFetcher(CompositeTask):
    __slots__ = (
        "config_pool",
        "ebuild_path",
        "fetchonly",
        "fetchall",
        "logfile",
        "pkg",
        "prefetch",
        "pre_exec",
        "_fetcher_proc",
    )

    def __init__(self, **kwargs):
        CompositeTask.__init__(self, **kwargs)
        self._fetcher_proc = _EbuildFetcherProcess(**kwargs)

    def async_already_fetched(self, settings):
        """
        Returns True if all files already exist locally and have correct
        digests, otherwise return False. When returning True, appropriate
        digest checking messages are produced for display and/or logging.
        When returning False, no messages are produced, since we assume
        that a fetcher process will later be executed in order to produce
        such messages. This will raise InvalidDependString if SRC_URI is
        invalid.
        """
        return self._fetcher_proc.async_already_fetched(settings)

    def _start(self):
        self._start_task(
            AsyncTaskFuture(future=self._fetcher_proc._async_uri_map()),
            self._start_fetch,
        )

    def _start_fetch(self, uri_map_task):
        self._assert_current(uri_map_task)
        if uri_map_task.cancelled:
            self._default_final_exit(uri_map_task)
            return

        try:
            uri_map = uri_map_task.future.result()
        except portage.exception.InvalidDependString as e:
            msg_lines = []
            msg = f"Fetch failed for '{self.pkg.cpv}' due to invalid SRC_URI: {e}"
            msg_lines.append(msg)
            self._fetcher_proc._eerror(msg_lines)
            self._current_task = None
            self.returncode = 1
            self._async_wait()
            return

        # First get the SRC_URI metadata (it's not cached in self.pkg.metadata
        # because some packages have an extremely large SRC_URI value).
        self._start_task(
            AsyncTaskFuture(
                future=self.pkg.root_config.trees["porttree"].dbapi.async_aux_get(
                    self.pkg.cpv, ["SRC_URI"], myrepo=self.pkg.repo, loop=self.scheduler
                )
            ),
            self._start_with_metadata,
        )

    def _start_with_metadata(self, aux_get_task):
        self._assert_current(aux_get_task)
        if aux_get_task.cancelled:
            self._default_final_exit(aux_get_task)
            return

        (self._fetcher_proc.src_uri,) = aux_get_task.future.result()
        self._start_task(self._fetcher_proc, self._default_final_exit)


class _EbuildFetcherProcess(ForkProcess):
    __slots__ = (
        "config_pool",
        "ebuild_path",
        "fetchonly",
        "fetchall",
        "pkg",
        "prefetch",
        "src_uri",
        "_digests",
        "_manifest",
        "_settings",
        "_uri_map",
    )

    def async_already_fetched(self, settings):
        result = self.scheduler.create_future()

        def uri_map_done(uri_map_future):
            if uri_map_future.cancelled():
                result.cancel()
                return

            if uri_map_future.exception() is not None or result.cancelled():
                if not result.cancelled():
                    result.set_exception(uri_map_future.exception())
                return

            uri_map = uri_map_future.result()
            if uri_map:
                result.set_result(self._check_already_fetched(settings, uri_map))
            else:
                result.set_result(True)

        uri_map_future = self._async_uri_map()
        result.add_done_callback(
            lambda result: uri_map_future.cancel() if result.cancelled() else None
        )
        uri_map_future.add_done_callback(uri_map_done)
        return result

    def _check_already_fetched(self, settings, uri_map):
        digests = self._get_digests()
        distdir = settings["DISTDIR"]
        allow_missing = self._get_manifest().allow_missing

        for filename in uri_map:
            # Use stat rather than lstat since fetch() creates
            # symlinks when PORTAGE_RO_DISTDIRS is used.
            try:
                st = os.stat(os.path.join(distdir, filename))
            except OSError:
                return False
            if st.st_size == 0:
                return False
            expected_size = digests.get(filename, {}).get("size")
            if expected_size is None:
                continue
            if st.st_size != expected_size:
                return False

        hash_filter = _hash_filter(settings.get("PORTAGE_CHECKSUM_FILTER", ""))
        if hash_filter.transparent:
            hash_filter = None
        stdout_orig = sys.stdout
        stderr_orig = sys.stderr
        global_havecolor = portage.output.havecolor
        out = io.StringIO()
        eout = portage.output.EOutput()
        eout.quiet = settings.get("PORTAGE_QUIET") == "1"
        success = True
        try:
            sys.stdout = out
            sys.stderr = out
            if portage.output.havecolor:
                portage.output.havecolor = not self.background

            for filename in uri_map:
                mydigests = digests.get(filename)
                if mydigests is None:
                    if not allow_missing:
                        success = False
                        break
                    continue
                ok, st = _check_distfile(
                    os.path.join(distdir, filename),
                    mydigests,
                    eout,
                    show_errors=False,
                    hash_filter=hash_filter,
                )
                if not ok:
                    success = False
                    break
        except portage.exception.FileNotFound:
            # A file disappeared unexpectedly.
            return False
        finally:
            sys.stdout = stdout_orig
            sys.stderr = stderr_orig
            portage.output.havecolor = global_havecolor

        if success:
            # When returning unsuccessfully, no messages are produced, since
            # we assume that a fetcher process will later be executed in order
            # to produce such messages.
            msg = out.getvalue()
            if msg:
                self.scheduler.output(msg, log_path=self.logfile)

        return success

    def _start(self):
        root_config = self.pkg.root_config
        portdb = root_config.trees["porttree"].dbapi
        ebuild_path = self._get_ebuild_path()
        # This is initialized by an earlier _async_uri_map call.
        uri_map = self._uri_map

        if not uri_map:
            # Nothing to fetch.
            self.returncode = os.EX_OK
            self._async_wait()
            return

        settings = self.config_pool.allocate()
        settings.setcpv(self.pkg)
        settings.configdict["pkg"]["SRC_URI"] = self.src_uri
        portage.doebuild_environment(ebuild_path, "fetch", settings=settings, db=portdb)

        if self.prefetch and self._prefetch_size_ok(uri_map, settings, ebuild_path):
            self.config_pool.deallocate(settings)
            self.returncode = os.EX_OK
            self._async_wait()
            return

        nocolor = settings.get("NOCOLOR")

        if self.prefetch:
            settings["PORTAGE_PARALLEL_FETCHONLY"] = "1"

        if self.background:
            nocolor = "true"

        if nocolor is not None:
            settings["NOCOLOR"] = nocolor

        self._settings = settings
        self.log_filter_file = settings.get("PORTAGE_LOG_FILTER_FILE_CMD")
        self.target = functools.partial(
            self._target,
            self._settings,
            self._get_manifest(),
            self._uri_map,
            self.fetchonly,
            self.pre_exec,
        )
        ForkProcess._start(self)

        # Free settings now since it's no longer needed in
        # this process (the subprocess has a private copy).
        self.config_pool.deallocate(settings)
        settings = None
        self._settings = None

    @staticmethod
    def _target(settings, manifest, uri_map, fetchonly, pre_exec):
        if pre_exec is not None:
            pre_exec()

        # Force consistent color output, in case we are capturing fetch
        # output through a normal pipe due to unavailability of ptys.
        portage.output.havecolor = settings.get("NOCOLOR") not in ("yes", "true")

        # For userfetch, drop privileges for the entire fetch call, in
        # order to handle DISTDIR on NFS with root_squash for bug 601252.
        if _want_userfetch(settings):
            _drop_privs_userfetch(settings)

        allow_missing = manifest.allow_missing or "digest" in settings.features

        async def main():
            loop = asyncio.get_event_loop()
            task = asyncio.ensure_future(
                async_fetch(
                    uri_map,
                    settings,
                    fetchonly=fetchonly,
                    digests=copy.deepcopy(manifest.getTypeDigests("DIST")),
                    allow_missing_digests=allow_missing,
                )
            )

            def sigterm_handler(signum, _frame):
                loop.call_soon_threadsafe(task.cancel)
                signal.signal(signal.SIGTERM, signal.SIG_IGN)

            signal.signal(signal.SIGTERM, sigterm_handler)
            try:
                await task
            except asyncio.CancelledError:
                # If asyncio.CancelledError arrives too soon after fork/spawn
                # then handers will not have an opportunity to terminate
                # the corresponding process, so clean up after this race.
                for proc in multiprocessing.active_children():
                    proc.terminate()

                # Use a non-zero timeout only for the first join because
                # later joins are delayed by the first join.
                timeout = 0.25
                for proc in multiprocessing.active_children():
                    proc.join(timeout)
                    timeout = 0

                for proc in multiprocessing.active_children():
                    proc.kill()
                    # Wait upon the process in order to ensure that its
                    # pid will trigger ProcessLookupError for tests.
                    proc.join()

                signal.signal(signal.SIGTERM, signal.SIG_DFL)
                os.kill(os.getpid(), signal.SIGTERM)

            return os.EX_OK if task.result() else 1

        return asyncio.run(main())

    def _get_ebuild_path(self):
        if self.ebuild_path is not None:
            return self.ebuild_path
        portdb = self.pkg.root_config.trees["porttree"].dbapi
        self.ebuild_path = portdb.findname(self.pkg.cpv, myrepo=self.pkg.repo)
        if self.ebuild_path is None:
            raise AssertionError(f"ebuild not found for '{self.pkg.cpv}'")
        return self.ebuild_path

    def _get_manifest(self):
        if self._manifest is None:
            pkgdir = os.path.dirname(self._get_ebuild_path())
            self._manifest = (
                self.pkg.root_config.settings.repositories.get_repo_for_location(
                    os.path.dirname(os.path.dirname(pkgdir))
                ).load_manifest(pkgdir, None)
            )
        return self._manifest

    def _get_digests(self):
        if self._digests is None:
            self._digests = self._get_manifest().getTypeDigests("DIST")
        return self._digests

    def _async_uri_map(self):
        """
        This calls the portdbapi.async_fetch_map method and returns the
        resulting Future (may contain InvalidDependString exception).
        """
        if self._uri_map is not None:
            result = self.scheduler.create_future()
            result.set_result(self._uri_map)
            return result

        pkgdir = os.path.dirname(self._get_ebuild_path())
        mytree = os.path.dirname(os.path.dirname(pkgdir))
        use = None
        if not self.fetchall:
            use = self.pkg.use.enabled
        portdb = self.pkg.root_config.trees["porttree"].dbapi

        def cache_result(result):
            try:
                self._uri_map = result.result()
            except (asyncio.CancelledError, Exception):
                # The caller handles this when it retrieves the result.
                pass

        result = portdb.async_fetch_map(
            self.pkg.cpv, useflags=use, mytree=mytree, loop=self.scheduler
        )
        result.add_done_callback(cache_result)
        return result

    def _prefetch_size_ok(self, uri_map, settings, ebuild_path):
        distdir = settings["DISTDIR"]

        sizes = {}
        for filename in uri_map:
            # Use stat rather than lstat since portage.fetch() creates
            # symlinks when PORTAGE_RO_DISTDIRS is used.
            try:
                st = os.stat(os.path.join(distdir, filename))
            except OSError:
                return False
            if st.st_size == 0:
                return False
            sizes[filename] = st.st_size

        digests = self._get_digests()
        for filename, actual_size in sizes.items():
            size = digests.get(filename, {}).get("size")
            if size is None:
                continue
            if size != actual_size:
                return False

        # All files are present and sizes are ok. In this case the normal
        # fetch code will be skipped, so we need to generate equivalent
        # output here.
        if self.logfile is not None:
            f = open(
                _unicode_encode(
                    self.logfile, encoding=_encodings["fs"], errors="strict"
                ),
                mode="a",
                encoding=_encodings["content"],
                errors="backslashreplace",
            )
            for filename in uri_map:
                f.write(
                    _unicode_decode(
                        f" * {filename} size ;-) ...".ljust(73) + "[ ok ]\n"
                    )
                )
            f.close()

        return True

    def _pipe(self, fd_pipes):
        """When appropriate, use a pty so that fetcher progress bars,
        like wget has, will work properly."""
        if self.background or not sys.stdout.isatty():
            # When the output only goes to a log file,
            # there's no point in creating a pty.
            return os.pipe()
        stdout_pipe = None
        if not self.background:
            stdout_pipe = fd_pipes.get(1)
        self._pty_ready, master_fd, slave_fd = _create_pty_or_pipe(
            copy_term_size=stdout_pipe
        )
        return (master_fd, slave_fd)

    def _eerror(self, lines):
        out = io.StringIO()
        for line in lines:
            eerror(line, phase="unpack", key=self.pkg.cpv, out=out)
        msg = out.getvalue()
        if msg:
            self.scheduler.output(msg, log_path=self.logfile)

    def _proc_join_done(self, proc, future):
        """
        Extend _proc_join_done to emit an eerror message for fetch failure.
        """
        if not self.prefetch and not future.cancelled() and proc.exitcode != os.EX_OK:
            msg_lines = []
            msg = f"Fetch failed for '{self.pkg.cpv}'"
            if self.logfile is not None:
                msg += ", Log file:"
            msg_lines.append(msg)
            if self.logfile is not None:
                msg_lines.append(f" '{self.logfile}'")
            self._eerror(msg_lines)
        super()._proc_join_done(proc, future)



# Copyright 2012-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import fcntl
import multiprocessing
import warnings
import signal
import sys

from typing import Optional

import portage
from portage import os
from portage.cache.mappings import slot_dict_class
from portage.util.futures import asyncio
from _emerge.SpawnProcess import SpawnProcess

_registered_run_exitfuncs = None


class ForkProcess(SpawnProcess):
    # NOTE: This class overrides the meaning of the SpawnProcess 'args'
    # attribute, and uses it to hold the positional arguments for the
    # 'target' function.
    __slots__ = (
        "kwargs",
        "target",
        "_child_connection",
        # Duplicate file descriptors for use by _send_fd_pipes background thread.
        "_fd_pipes",
    )

    _file_names = ("connection", "slave_fd")
    _files_dict = slot_dict_class(_file_names, prefix="")

    _HAVE_SEND_HANDLE = getattr(multiprocessing.reduction, "HAVE_SEND_HANDLE", False)

    def _start(self):
        if multiprocessing.get_start_method() == "fork":
            # Backward compatibility mode.
            super()._start()
            return

        # This mode supports multiprocessing start methods
        # other than fork. Note that the fd_pipes implementation
        # uses a thread via run_in_executor, and threads are not
        # recommended for mixing with the fork start method due
        # to cpython issue 84559.
        if self.fd_pipes and not self._HAVE_SEND_HANDLE:
            raise NotImplementedError(
                'fd_pipes only supported with HAVE_SEND_HANDLE or multiprocessing start method "fork"'
            )

        if self.fd_pipes or self.logfile or not self.background:
            # Log via multiprocessing.Pipe if necessary.
            connection, self._child_connection = multiprocessing.Pipe(
                duplex=self._HAVE_SEND_HANDLE
            )

        # Handle fd_pipes in _main instead, since file descriptors are
        # not inherited with the multiprocessing "spawn" start method.
        # Pass fd_pipes=None to spawn here so that it doesn't leave
        # a closed stdin duplicate in fd_pipes (that would trigger
        # "Bad file descriptor" error if we tried to send it via
        # send_handle).
        self._proc = self._spawn(self.args, fd_pipes=None)

        self._registered = True

        if self._child_connection is None:
            self._async_waitpid()
        else:
            self._child_connection.close()
            self.fd_pipes = self.fd_pipes or {}
            stdout_fd = None
            if not self.background:
                self.fd_pipes.setdefault(0, portage._get_stdin().fileno())
                self.fd_pipes.setdefault(1, sys.__stdout__.fileno())
                self.fd_pipes.setdefault(2, sys.__stderr__.fileno())
                if self.create_pipe is not False:
                    stdout_fd = os.dup(self.fd_pipes[1])

            if self._HAVE_SEND_HANDLE:
                if self.create_pipe is not False:
                    master_fd, slave_fd = self._pipe(self.fd_pipes)
                    self.fd_pipes[1] = slave_fd
                    self.fd_pipes[2] = slave_fd
                else:
                    if self.logfile:
                        raise NotImplementedError(
                            "logfile conflicts with create_pipe=False"
                        )
                    # When called via process.spawn, SpawnProcess
                    # will have created a pipe earlier, so it would be
                    # redundant to do it here (it could also trigger spawn
                    # recursion via set_term_size as in bug 923750).
                    master_fd = None
                    slave_fd = None

                self._files = self._files_dict(connection=connection, slave_fd=slave_fd)

                # Create duplicate file descriptors in self._fd_pipes
                # so that the caller is free to manage the lifecycle
                # of the original fd_pipes.
                self._fd_pipes = {}
                fd_map = {}
                for dest, src in list(self.fd_pipes.items()):
                    if src not in fd_map:
                        src_new = fd_map[src] = os.dup(src)
                        old_fdflags = fcntl.fcntl(src, fcntl.F_GETFD)
                        fcntl.fcntl(src_new, fcntl.F_SETFD, old_fdflags)
                        os.set_inheritable(
                            src_new, not bool(old_fdflags & fcntl.FD_CLOEXEC)
                        )
                    self._fd_pipes[dest] = fd_map[src]

                asyncio.ensure_future(
                    self._proc.wait(), self.scheduler
                ).add_done_callback(self._close_fd_pipes)
            else:
                master_fd = connection

            self._start_main_task(
                master_fd, log_file_path=self.logfile, stdout_fd=stdout_fd
            )

    def _close_fd_pipes(self, future):
        """
        Cleanup self._fd_pipes if needed, since _send_fd_pipes could
        have been cancelled.
        """
        # future.result() raises asyncio.CancelledError if
        # future.cancelled(), but that should not happen.
        future.result()
        if self._fd_pipes is not None:
            for fd in set(self._fd_pipes.values()):
                os.close(fd)
            self._fd_pipes = None

    @property
    def _fd_pipes_send_handle(self):
        """Returns True if we have a connection to implement fd_pipes via send_handle."""
        return bool(
            self._HAVE_SEND_HANDLE
            and self._files
            and getattr(self._files, "connection", False)
        )

    def _send_fd_pipes(self):
        """
        Communicate with _bootstrap to send fd_pipes via send_handle.
        This performs blocking IO, intended for invocation via run_in_executor.
        """
        fd_list = list(set(self._fd_pipes.values()))
        try:
            self._files.connection.send(
                (self._fd_pipes, fd_list),
            )
            for fd in fd_list:
                multiprocessing.reduction.send_handle(
                    self._files.connection,
                    fd,
                    self.pid,
                )
        except BrokenPipeError as e:
            # This case is triggered by testAsynchronousLockWaitCancel
            # when the test case terminates the child process while
            # this thread is still sending the fd_pipes (bug 923852).
            # Even if the child terminated abnormally, then there is
            # no harm in suppressing the exception here, since the
            # child error should have gone to stderr.
            raise asyncio.CancelledError from e

        # self._fd_pipes contains duplicates that must be closed.
        for fd in fd_list:
            os.close(fd)
        self._fd_pipes = None

    async def _main(self, build_logger, pipe_logger, loop=None):
        try:
            if self._fd_pipes_send_handle:
                await self.scheduler.run_in_executor(
                    None,
                    self._send_fd_pipes,
                )
        except asyncio.CancelledError:
            self._main_cancel(build_logger, pipe_logger)
            raise
        finally:
            if self._files:
                if hasattr(self._files, "connection"):
                    self._files.connection.close()
                    del self._files.connection
                if hasattr(self._files, "slave_fd"):
                    if self._files.slave_fd is not None:
                        os.close(self._files.slave_fd)
                    del self._files.slave_fd

        await super()._main(build_logger, pipe_logger, loop=loop)

    def _spawn(
        self, args: list[str], fd_pipes: Optional[dict[int, int]] = None, **kwargs
    ) -> portage.process.MultiprocessingProcess:
        """
        Override SpawnProcess._spawn to fork a subprocess that calls
        self._run(). This uses multiprocessing.Process in order to leverage
        any pre-fork and post-fork interpreter housekeeping that it provides,
        promoting a healthy state for the forked interpreter.
        """

        global _registered_run_exitfuncs

        if self.__class__._run is ForkProcess._run:
            # target replaces the deprecated self._run method
            target = self.target
            args = self.args
            kwargs = self.kwargs
        else:
            # _run implementation triggers backward-compatibility mode
            target = self._run
            args = None
            kwargs = None
            warnings.warn(
                'portage.util._async.ForkProcess.ForkProcess._run is deprecated in favor of the "target" parameter',
                UserWarning,
                stacklevel=2,
            )

        # Since multiprocessing.Process closes sys.__stdin__, create a
        # temporary duplicate of fd_pipes[0] so that sys.__stdin__ can
        # be restored in the subprocess, in case this is needed for
        # things like PROPERTIES=interactive support.
        stdin_dup = None
        try:
            stdin_fd = fd_pipes.get(0) if fd_pipes else None
            if stdin_fd is not None and stdin_fd == portage._get_stdin().fileno():
                stdin_dup = os.dup(stdin_fd)
                fcntl.fcntl(
                    stdin_dup, fcntl.F_SETFD, fcntl.fcntl(stdin_fd, fcntl.F_GETFD)
                )
                fd_pipes[0] = stdin_dup

            proc = multiprocessing.Process(
                target=self._bootstrap,
                args=(
                    self._child_connection,
                    self._HAVE_SEND_HANDLE,
                    fd_pipes,
                    target,
                    args,
                    kwargs,
                ),
            )
            proc.start()
        finally:
            if stdin_dup is not None:
                os.close(stdin_dup)

        if _registered_run_exitfuncs != portage.getpid():
            portage.process._atexit_register_run_exitfuncs()
            _registered_run_exitfuncs = portage.getpid()

        return portage.process.MultiprocessingProcess(proc)

    def _cancel(self):
        if self._proc is None:
            super()._cancel()
        else:
            self._proc.terminate()

    def _unregister(self):
        super()._unregister()
        if self._proc is not None:
            self._proc.terminate()

    @staticmethod
    def _bootstrap(child_connection, have_send_handle, fd_pipes, target, args, kwargs):
        # Use default signal handlers in order to avoid problems
        # killing subprocesses as reported in bug #353239.
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        # Unregister SIGCHLD handler and wakeup_fd for the parent
        # process's event loop (bug 655656).
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        try:
            wakeup_fd = signal.set_wakeup_fd(-1)
            if wakeup_fd > 0:
                os.close(wakeup_fd)
        except (ValueError, OSError):
            pass

        portage.locks._close_fds()

        if child_connection is not None:
            if have_send_handle:
                fd_pipes, fd_list = child_connection.recv()
                fd_pipes_map = {}
                for fd in fd_list:
                    fd_pipes_map[fd] = multiprocessing.reduction.recv_handle(
                        child_connection
                    )
                child_connection.close()
                for k, v in list(fd_pipes.items()):
                    fd_pipes[k] = fd_pipes_map[v]

            else:
                fd_pipes = fd_pipes or {}
                fd_pipes[sys.stdout.fileno()] = child_connection.fileno()
                fd_pipes[sys.stderr.fileno()] = child_connection.fileno()
                fd_pipes[child_connection.fileno()] = child_connection.fileno()

        if fd_pipes:
            # We don't exec, so use close_fds=False
            # (see _setup_pipes docstring).
            portage.process._setup_pipes(fd_pipes, close_fds=False)

        # Since multiprocessing.Process closes sys.__stdin__ and
        # makes sys.stdin refer to os.devnull, restore it when
        # appropriate.
        if fd_pipes and 0 in fd_pipes:
            # It's possible that sys.stdin.fileno() is already 0,
            # and in that case the above _setup_pipes call will
            # have already updated its identity via dup2. Otherwise,
            # perform the dup2 call now, and also copy the file
            # descriptor flags.
            if sys.stdin.fileno() != 0:
                os.dup2(0, sys.stdin.fileno())
                fcntl.fcntl(
                    sys.stdin.fileno(), fcntl.F_SETFD, fcntl.fcntl(0, fcntl.F_GETFD)
                )
            sys.__stdin__ = sys.stdin

        sys.exit(target(*(args or []), **(kwargs or {})))

    def _run(self):
        """
        Deprecated and replaced with the "target" constructor parameter.
        """
        raise NotImplementedError(self)



# Copyright 2005-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# Author(s): Brian Harring (ferringb@gentoo.org)

from portage.cache import fs_template
from portage.cache import cache_errors
import errno
import stat
import tempfile
import os as _os
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage.exception import InvalidData
from portage.versions import _pkg_str


class database(fs_template.FsBased):
    autocommits = True

    def __init__(self, *args, **config):
        super().__init__(*args, **config)
        self.location = os.path.join(
            self.location, self.label.lstrip(os.path.sep).rstrip(os.path.sep)
        )
        write_keys = set(self._known_keys)
        write_keys.add("_eclasses_")
        write_keys.add(f"_{self.validation_chf}_")
        self._write_keys = sorted(write_keys)
        if not self.readonly and not os.path.exists(self.location):
            self._ensure_dirs()

    def _getitem(self, cpv):
        # Don't use os.path.join, for better performance.
        fp = self.location + _os.sep + cpv
        try:
            with open(
                _unicode_encode(fp, encoding=_encodings["fs"], errors="strict"),
                encoding=_encodings["repo.content"],
                errors="replace",
            ) as myf:
                lines = myf.read().split("\n")
                if not lines[-1]:
                    lines.pop()
                d = self._parse_data(lines, cpv)
                if "_mtime_" not in d:
                    # Backward compatibility with old cache
                    # that uses mtime mangling.
                    d["_mtime_"] = _os.fstat(myf.fileno())[stat.ST_MTIME]
                return d
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise cache_errors.CacheCorruption(cpv, e)
            raise KeyError(cpv, e)

    def _parse_data(self, data, cpv):
        try:
            return dict(x.split("=", 1) for x in data)
        except ValueError as e:
            # If a line is missing an "=", the split length is 1 instead of 2.
            raise cache_errors.CacheCorruption(cpv, e)

    def _setitem(self, cpv, values):
        try:
            fd, fp = tempfile.mkstemp(dir=self.location)
        except OSError as e:
            raise cache_errors.CacheCorruption(cpv, e)

        with open(
            fd, mode="w", encoding=_encodings["repo.content"], errors="backslashreplace"
        ) as myf:
            for k in self._write_keys:
                v = values.get(k)
                if not v:
                    continue
                myf.write(f"{k}={v}\n")

        self._ensure_access(fp)

        # update written.  now we move it.

        new_fp = os.path.join(self.location, cpv)
        try:
            os.rename(fp, new_fp)
        except OSError as e:
            success = False
            try:
                if errno.ENOENT == e.errno:
                    try:
                        self._ensure_dirs(cpv)
                        os.rename(fp, new_fp)
                        success = True
                    except OSError as e:
                        raise cache_errors.CacheCorruption(cpv, e)
                else:
                    raise cache_errors.CacheCorruption(cpv, e)
            finally:
                if not success:
                    os.remove(fp)

    def _delitem(self, cpv):
        # 		import pdb;pdb.set_trace()
        try:
            os.remove(os.path.join(self.location, cpv))
        except OSError as e:
            if errno.ENOENT == e.errno:
                raise KeyError(cpv)
            else:
                raise cache_errors.CacheCorruption(cpv, e)

    def __contains__(self, cpv):
        return os.path.exists(os.path.join(self.location, cpv))

    def __iter__(self):
        """generator for walking the dir struct"""
        dirs = [(0, self.location)]
        len_base = len(self.location)
        while dirs:
            depth, dir_path = dirs.pop()
            try:
                dir_list = os.listdir(dir_path)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
                del e
                continue
            for l in dir_list:
                p = os.path.join(dir_path, l)
                try:
                    st = os.lstat(p)
                except OSError:
                    # Cache entry disappeared.
                    continue
                if stat.S_ISDIR(st.st_mode):
                    # Only recurse 1 deep, in order to avoid iteration over
                    # entries from another nested cache instance. This can
                    # happen if the user nests an overlay inside
                    # /var/db/repos/gentoo/local as in bug #302764.
                    if depth < 1:
                        dirs.append((depth + 1, p))
                    continue

                try:
                    yield _pkg_str(p[len_base + 1 :])
                except InvalidData:
                    continue


class md5_database(database):
    validation_chf = "md5"
    store_eclass_paths = False


class mtime_md5_database(database):
    validation_chf = "md5"
    chf_types = ("md5", "mtime")



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tempfile

from portage import os
from portage import shutil
from portage.const import EBUILD_PHASES
from portage.elog import elog_process
from portage.package.ebuild.config import config
from portage.package.ebuild.doebuild import doebuild_environment
from portage.package.ebuild.prepare_build_dirs import prepare_build_dirs
from portage.util.futures import asyncio
from portage.util._async.SchedulerInterface import SchedulerInterface
from _emerge.CompositeTask import CompositeTask
from _emerge.EbuildPhase import EbuildPhase


class SpawnNofetchWithoutBuilddir(CompositeTask):
    """
    This spawns pkg_nofetch if appropriate, while avoiding the
    need to lock a global build directory. The settings parameter
    is useful only if setcpv has already been called in order
    to cache metadata. It will be cloned internally, in order to
    prevent any changes from interfering with the calling code.
    If settings is None then a suitable config instance will be
    acquired from the given portdbapi instance. Do not use the
    settings parameter unless setcpv has been called on the given
    instance, since otherwise it's possible to trigger issues like
    bug #408817 due to fragile assumptions involving the config
    state inside doebuild_environment().

    A private PORTAGE_BUILDDIR will be created and cleaned up, in
    order to avoid any interference with any other processes.
    If PORTAGE_TMPDIR is writable, that will be used, otherwise
    the default directory for the tempfile module will be used.

    We only call the pkg_nofetch phase if either RESTRICT=fetch
    is set or the package has explicitly overridden the default
    pkg_nofetch implementation. This allows specialized messages
    to be displayed for problematic packages even though they do
    not set RESTRICT=fetch (bug #336499).

    This class does nothing if the PORTAGE_PARALLEL_FETCHONLY
    variable is set in the config instance.
    """

    __slots__ = ("ebuild_path", "fd_pipes", "portdb", "settings", "_private_tmpdir")

    def _start(self):
        settings = self.settings
        if settings is None:
            settings = self.portdb.settings

        if "PORTAGE_PARALLEL_FETCHONLY" in settings:
            # parallel-fetch mode
            self.returncode = os.EX_OK
            self._async_wait()
            return

        # Prevent temporary config changes from interfering
        # with config instances that are reused.
        settings = self.settings = config(clone=settings)

        # We must create our private PORTAGE_TMPDIR before calling
        # doebuild_environment(), since lots of variables such
        # as PORTAGE_BUILDDIR refer to paths inside PORTAGE_TMPDIR.
        portage_tmpdir = settings.get("PORTAGE_TMPDIR")
        if not portage_tmpdir or not os.access(portage_tmpdir, os.W_OK):
            portage_tmpdir = None
        private_tmpdir = self._private_tmpdir = tempfile.mkdtemp(dir=portage_tmpdir)
        settings["PORTAGE_TMPDIR"] = private_tmpdir
        settings.backup_changes("PORTAGE_TMPDIR")
        # private temp dir was just created, so it's not locked yet
        settings.pop("PORTAGE_BUILDDIR_LOCKED", None)

        doebuild_environment(
            self.ebuild_path, "nofetch", settings=settings, db=self.portdb
        )
        restrict = settings["PORTAGE_RESTRICT"].split()
        defined_phases = settings["DEFINED_PHASES"].split()
        if not defined_phases:
            # When DEFINED_PHASES is undefined, assume all
            # phases are defined.
            defined_phases = EBUILD_PHASES

        if "fetch" not in restrict and "nofetch" not in defined_phases:
            self.returncode = os.EX_OK
            self._async_wait()
            return

        prepare_build_dirs(settings=settings)

        ebuild_phase = EbuildPhase(
            background=self.background,
            phase="nofetch",
            scheduler=self.scheduler,
            fd_pipes=self.fd_pipes,
            settings=settings,
        )

        self._start_task(ebuild_phase, self._nofetch_exit)

    def _nofetch_exit(self, ebuild_phase):
        self._final_exit(ebuild_phase)
        elog_process(self.settings.mycpv, self.settings)
        shutil.rmtree(self._private_tmpdir)
        self._async_wait()


def spawn_nofetch(portdb, ebuild_path, settings=None, fd_pipes=None):
    """
    Create a NofetchPrivateTmpdir instance, and execute it synchronously.
    This function must not be called from asynchronous code, since it will
    trigger event loop recursion which is incompatible with asyncio.
    """
    nofetch = SpawnNofetchWithoutBuilddir(
        background=False,
        portdb=portdb,
        ebuild_path=ebuild_path,
        scheduler=SchedulerInterface(asyncio._safe_loop()),
        fd_pipes=fd_pipes,
        settings=settings,
    )

    nofetch.start()
    return nofetch.wait()



# Copyright 2010-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import signal
import sys
import tempfile

import pytest

from portage import os
from portage import shutil
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from _emerge.AsynchronousLock import AsynchronousLock


class AsynchronousLockTestCase(TestCase):
    def _testAsynchronousLock(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            for force_async in (True, False):
                async_lock = AsynchronousLock(
                    path=path,
                    scheduler=scheduler,
                    _force_async=force_async,
                    _force_thread=True,
                )
                async_lock.start()
                self.assertEqual(async_lock.wait(), os.EX_OK)
                self.assertEqual(async_lock.returncode, os.EX_OK)
                scheduler.run_until_complete(async_lock.async_unlock())

                async_lock = AsynchronousLock(
                    path=path,
                    scheduler=scheduler,
                    _force_async=force_async,
                    _force_process=True,
                )
                async_lock.start()
                self.assertEqual(async_lock.wait(), os.EX_OK)
                self.assertEqual(async_lock.returncode, os.EX_OK)
                scheduler.run_until_complete(async_lock.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    def testAsynchronousLock(self):
        self._testAsynchronousLock()

    def testAsynchronousLockHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLock()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def _testAsynchronousLockWait(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = AsynchronousLock(path=path, scheduler=scheduler)
            lock1.start()
            self.assertEqual(lock1.wait(), os.EX_OK)
            self.assertEqual(lock1.returncode, os.EX_OK)

            # lock2 requires _force_async=True since the portage.locks
            # module is not designed to work as intended here if the
            # same process tries to lock the same file more than
            # one time concurrently.
            lock2 = AsynchronousLock(
                path=path, scheduler=scheduler, _force_async=True, _force_process=True
            )
            lock2.start()
            # lock2 should be waiting for lock1 to release
            self.assertEqual(lock2.poll(), None)
            self.assertEqual(lock2.returncode, None)

            scheduler.run_until_complete(lock1.async_unlock())
            self.assertEqual(lock2.wait(), os.EX_OK)
            self.assertEqual(lock2.returncode, os.EX_OK)
            scheduler.run_until_complete(lock2.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    def testAsynchronousLockWait(self):
        self._testAsynchronousLockWait()

    def testAsynchronousLockWaitHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLockWait()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def _testAsynchronousLockWaitCancel(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = AsynchronousLock(path=path, scheduler=scheduler)
            lock1.start()
            self.assertEqual(lock1.wait(), os.EX_OK)
            self.assertEqual(lock1.returncode, os.EX_OK)
            lock2 = AsynchronousLock(
                path=path, scheduler=scheduler, _force_async=True, _force_process=True
            )
            lock2.start()
            # lock2 should be waiting for lock1 to release
            self.assertEqual(lock2.poll(), None)
            self.assertEqual(lock2.returncode, None)

            # Cancel lock2 and then check wait() and returncode results.
            lock2.cancel()
            self.assertEqual(lock2.wait() == os.EX_OK, False)
            self.assertEqual(lock2.returncode == os.EX_OK, False)
            self.assertEqual(lock2.returncode is None, False)
            scheduler.run_until_complete(lock1.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    @pytest.mark.skipif(
        sys.version_info >= (3, 14), reason="fails with python 3.14.0a3"
    )
    def testAsynchronousLockWaitCancel(self):
        self._testAsynchronousLockWaitCancel()

    @pytest.mark.skipif(
        sys.version_info >= (3, 14), reason="fails with python 3.14.0a3"
    )
    def testAsynchronousLockWaitCancelHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLockWaitCancel()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def _testAsynchronousLockWaitKill(self):
        scheduler = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = AsynchronousLock(path=path, scheduler=scheduler)
            lock1.start()
            self.assertEqual(lock1.wait(), os.EX_OK)
            self.assertEqual(lock1.returncode, os.EX_OK)
            lock2 = AsynchronousLock(
                path=path, scheduler=scheduler, _force_async=True, _force_process=True
            )
            lock2.start()
            # lock2 should be waiting for lock1 to release
            self.assertEqual(lock2.poll(), None)
            self.assertEqual(lock2.returncode, None)

            # Kill lock2's process and then check wait() and
            # returncode results. This is intended to simulate
            # a SIGINT sent via the controlling tty.
            self.assertEqual(lock2._imp is not None, True)
            self.assertEqual(lock2._imp._proc is not None, True)
            self.assertEqual(lock2._imp._proc.pid is not None, True)
            lock2._imp._kill_test = True
            os.kill(lock2._imp._proc.pid, signal.SIGTERM)
            self.assertEqual(lock2.wait() == os.EX_OK, False)
            self.assertEqual(lock2.returncode == os.EX_OK, False)
            self.assertEqual(lock2.returncode is None, False)
            scheduler.run_until_complete(lock1.async_unlock())
        finally:
            shutil.rmtree(tempdir)

    def testAsynchronousLockWaitKill(self):
        self._testAsynchronousLockWaitKill()

    def testAsynchronousLockWaitKillHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testAsynchronousLockWaitKill()
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["cpv_expand"]

import portage
from portage.exception import AmbiguousPackageName
from portage.versions import _pkgsplit


def cpv_expand(mycpv, mydb=None, use_cache=1, settings=None):
    """Given a string (packagename or virtual) expand it into a valid
    cat/package string. Virtuals use the mydb to determine which provided
    virtual is a valid choice and defaults to the first element when there
    are no installed/available candidates."""
    myslash = mycpv.split("/")
    mysplit = _pkgsplit(myslash[-1])
    if settings is None:
        try:
            settings = mydb.settings
        except AttributeError:
            settings = portage.settings
    if len(myslash) > 2:
        # this is illegal case.
        mysplit = []
        mykey = mycpv
    elif len(myslash) == 2:
        if mysplit:
            mykey = myslash[0] + "/" + mysplit[0]
        else:
            mykey = mycpv

        # Since Gentoo stopped using old-style virtuals in
        # 2011, typically it's possible to avoid getvirtuals()
        # calls entirely. Therefore, only call getvirtuals()
        # if the atom category is "virtual" and cp_list()
        # returns nothing.
        if (
            mykey.startswith("virtual/")
            and hasattr(mydb, "cp_list")
            and not mydb.cp_list(mykey, use_cache=use_cache)
        ):
            if hasattr(mydb, "vartree"):
                settings._populate_treeVirtuals_if_needed(mydb.vartree)
            virts = settings.getvirtuals().get(mykey)
            if virts:
                mykey_orig = mykey
                for vkey in virts:
                    # The virtuals file can contain a versioned atom, so
                    # it may be necessary to remove the operator and
                    # version from the atom before it is passed into
                    # dbapi.cp_list().
                    if mydb.cp_list(vkey.cp):
                        mykey = str(vkey)
                        break
                if mykey == mykey_orig:
                    mykey = str(virts[0])
            # we only perform virtual expansion if we are passed a dbapi
    else:
        # specific cpv, no category, ie. "foo-1.0"
        if mysplit:
            myp = mysplit[0]
        else:
            # "foo" ?
            myp = mycpv
        mykey = None
        matches = []
        if mydb and hasattr(mydb, "categories"):
            for x in mydb.categories:
                if mydb.cp_list(x + "/" + myp, use_cache=use_cache):
                    matches.append(x + "/" + myp)
        if len(matches) > 1:
            virtual_name_collision = False
            if len(matches) > 1:
                for x in matches:
                    if not x.startswith(("acct-group/", "acct-user/", "virtual/")):
                        # Assume that the non-virtual is desired.  This helps
                        # avoid the ValueError for invalid deps that come from
                        # installed packages (during reverse blocker detection,
                        # for example).
                        mykey = x
                    else:
                        virtual_name_collision = True
            if not virtual_name_collision:
                # AmbiguousPackageName inherits from ValueError,
                # for backward compatibility with calling code
                # that already handles ValueError.
                raise AmbiguousPackageName(matches)
        elif matches:
            mykey = matches[0]

        if not mykey and not isinstance(mydb, list):
            if hasattr(mydb, "vartree"):
                settings._populate_treeVirtuals_if_needed(mydb.vartree)
            virts_p = settings.get_virts_p().get(myp)
            if virts_p:
                mykey = virts_p[0]
            # again, we only perform virtual expansion if we have a dbapi (not a list)
        if not mykey:
            mykey = "null/" + myp
    if mysplit:
        if mysplit[2] == "r0":
            return mykey + "-" + mysplit[1]
        return mykey + "-" + mysplit[1] + "-" + mysplit[2]
    return mykey



# Copyright 2005-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import time
import signal

import portage


class ProgressHandler:
    def __init__(self):
        self.reset()

    def reset(self):
        self.curval = 0
        self.maxval = 0
        self.last_update = 0
        self.min_display_latency = 0.2

    def onProgress(self, maxval, curval):
        self.maxval = maxval
        self.curval = curval
        cur_time = time.time()
        if cur_time - self.last_update >= self.min_display_latency:
            self.last_update = cur_time
            self.display()

    def display(self):
        raise NotImplementedError(self)


class ProgressBar(ProgressHandler):
    """Class to set up and return a Progress Bar"""

    def __init__(self, isatty, **kwargs):
        self.isatty = isatty
        self.kwargs = kwargs
        ProgressHandler.__init__(self)
        self.progressBar = None

    def start(self):
        if self.isatty:
            self.progressBar = portage.output.TermProgressBar(**self.kwargs)
            signal.signal(signal.SIGWINCH, self.sigwinch_handler)
        else:
            self.onProgress = None
        return self.onProgress

    def set_label(self, _label):
        self.kwargs["label"] = _label

    def display(self):
        self.progressBar.set(self.curval, self.maxval)

    def sigwinch_handler(self, signum, frame):
        lines, self.progressBar.term_columns = portage.output.get_term_size()

    def stop(self):
        signal.signal(signal.SIGWINCH, signal.SIG_DFL)
